#TODO: regression without constant

import pandas as pd
import numpy as np
import scipy as sp
from statsmodels.formula.api import ols, logit
from statsmodels.tools.tools import add_constant
from statsmodels.api import OLS, Logit
from statsmodels.stats.outliers_influence import OLSInfluence

from IPython.display import display

from pandas.api.types import is_numeric_dtype
from .utils import get_categories

class LinearRegression:
    
    """
    Class for OLS regression models based on the excellent statsmodels package.
    
    Parameters
    ----------
    method : 'enter' or 'backward' 
        Method for predictors selection
    include_constant : bool 
        (CURRENTLY UNAVAILIABLE) Whether to include constant in the model
    sig_level_entry : float 
        (CURRENTLY UNAVAILIABLE) Max significance level to include predictor in the model 
    sig_level_removal : float
        Min significance level to exclude predictor from the model

    Attributes
    ----------
    variables_excluded : list
        Variables excluded because of zero variance
    variables_included : list
        Variables included in a model
    predictions : pd.Series
        Predicted values
    N : int
        Number of observations included in a model
    r2 : float
        R-squared (coefficient of determination)
    r2_adjusted : float
        Adjusted r-squared
    F : float
        F-statistic
    F_pvalue : float
        P-value for F-statistic
    ess : float
        Explained sum of squares
    rss : float
        Residual sum of squares
    tss : float
        Total sum of squares
    coefficients : pd.Series
        Regression coefficients
    coefficients_sterrors : pd.Series
        Standard errors of regression coefficients 
    coefficients_tvalues : pd.Series
        T-statistics of regression coefficients 
    coefficients_pvalues : pd.Series
        P-values of regression coefficients
    """
    
    def __init__(
        self, 
        method='enter',
        include_constant=True,
        sig_level_entry=0.05,
        sig_level_removal=0.05
    ):
        
        self.method = method.lower().strip()
        self.include_constant = include_constant
        self.sig_level_entry = sig_level_entry
        self.sig_level_removal = sig_level_removal
        
    def fit(
        self,
        data,
        formula,
        categorical_variables=None,
        show_results=True,
        confidence_intervals=True,
        collinearity_statistics=False,
        use_patsy_notation=False,
        n_decimals=3
    ):
    
        """
        Fit model to the given data using formula.

        Parameters
        ----------
        data : pd.DataFrame 
            Data to fit a model  
        formula : str 
            Formula of a model specification, e.g. 'y ~ x1 + x2'; 
            should be passed either in Patsy (statsmodels) notation
            or using the following rules: 
            '*' for interaction of the variables,
            ':' for interaction & main effects, 
            i.e., 'y ~ x:z' equals to 'y ~ x + z + x*z' (unlike the Patsy notation).
            If you use Patsy notation, please specify the parameter use_patsy_notation=True.
        categorical_variables : list 
            List of names of the variables that should be considered categorical.
            These variables would be automatically converted into sets of dummy variables.
            If you want to use this option, please make sure that you don't have nested names of variables
            (e.g. 'imdb' and 'imdb_rate' at the same time), otherwise this option results in an incorrect procedure.
        show_results : bool 
            Whether to show results of analysis
        confidence_intervals : bool 
            Whether to include coefficients' confidence intervals in the summary table
        collinearity_statistics : bool 
            whether to include coefficients' tolerance and VIF in the summary table
        use_patsy_notation : bool 
            turn this on if you use strictly Patsy's rules to define a formula.
            See more: https://patsy.readthedocs.io/en/latest/quickstart.html
        n_decimals : int 
            Number of digits to round results when showing them

        Returns
        -------
        self
            The current instance of the LinearRegression class
        """

        self._data = data.copy()
        
        self.categorical_variables = categorical_variables
        self._show_ci = confidence_intervals
        self._show_col = collinearity_statistics
        
        if '=' in formula:
            formula = formula.replace('=', '~')
        
        if not use_patsy_notation:
            formula = formula.replace('*', '^').replace(':', '*').replace('^', ':')
            
        self.formula = formula
        
        #won't work correctly if some variables have similar names (e.g. kinopoisk_rate and kinopoisk_rate_count)
        if categorical_variables is not None:
            if not isinstance(categorical_variables, list):
                raise ValueError(f"""Categorical variables should be passed as list.
                Type {type(categorical_variables)} was passed instead.""")
            else:
                for variable in categorical_variables:
                    formula = formula.replace(variable, f'C({variable})')
                    
        self._model = ols(formula=formula, data=data).fit()
        self._observations_idx = list(self._model.fittedvalues.index)
        self.dependent_variable = self._model.model.endog_names        
        self.variables_excluded = self._identify_variables_without_variation()
        
        if len(self.variables_excluded) > 0:
            y = pd.Series(self._model.model.endog.copy(), 
                          index=self._observations_idx,
                          name=self.dependent_variable)
            X = self._remove_variables_without_variation()
            self._model = OLS(y, X, missing = 'drop').fit()
            self.variables_excluded = [LinearRegression._translate_from_patsy_notation(x) for x in self.variables_excluded]
        
        if self.method == 'backward':
            self._fit_backward()
        
        self._get_statistics_from_model()
        
        self.predictions = self.predict()
        
        if show_results:
            self.show_results(n_decimals)
        
        if len(self.variables_excluded) > 0:
            print('------------------\n')
            print(f"Following variables were excluded due to zero variance: {'; '.join(self.variables_excluded)}")
        
        return self
    
    def predict(
        self,
        data=None,
        add_to_data=False,
    ):
        """
        Predict values of a dependent variable for the given data using the fitted model.
        
        Parameters
        ----------
        data : pd.DataFrame 
            Data for predictions, 
            may be not specified if you want to predict values for the same data that were used to fit a model
        add_to_data : bool 
            Whether to merge predictions with the given data.
            Currently, this option returns data with a sorted index

        Returns
        -------
        pd.DataFrame
            Predictions
        """
        name = f'{self.dependent_variable} (predicted)'
        
        if data is None:
            data_init = self._data.copy()
            result = self._model.fittedvalues
            data_init[name] = result
            if add_to_data:
                return data_init
            else:
                return data_init[name].copy()
                  
        else:
            aux_model = ols(self.formula, data).fit()
            aux_data_idx = aux_model.fittedvalues.index
            aux_data_cols = aux_model.model.exog_names
            aux_data_cols = [LinearRegression._translate_from_patsy_notation(x)\
                              for x in aux_data_cols]
            aux_data = pd.DataFrame(aux_model.model.exog,
                                    index=aux_data_idx,
                                    columns=aux_data_cols)
            aux_X = add_constant(aux_data[self.variables_included].copy())
            aux_y = aux_model.model.endog.copy()
                  
            aux_model = OLS(aux_y, aux_X, missing='drop').fit()
            result = aux_model.fittedvalues
            result.name = name      
            if add_to_data:
                result = pd.concat([data, result], axis=1, sort=False)
                  
            return result
        
    
    def _get_statistics_from_model(self):
        
        self.N = self._model.nobs
        self.r2 = self._model.rsquared
        self.r2_adjusted = self._model.rsquared_adj
        self.F = self._model.fvalue
        self.F_pvalue = self._model.f_pvalue
        self.ess = self._model.ess
        self.rss = self._model.ssr
        if self.include_constant:
            self.tss = self._model.centered_tss
        else:
            self.tss = self._model.uncentered_tss
        self.ms_model = self._model.mse_model
        self.ms_resid = self._model.mse_resid
        self.ms_total = self._model.mse_total
        self.dof_model = self._model.df_model
        self.dof_resid = self._model.df_resid
        self.dof_total = self.dof_model + self.dof_resid
        
        self.coefficients = self._model.params.copy()
        self.coefficients_sterrors = self._model.bse.copy()
        self.coefficients_tvalues = self._model.tvalues.copy()
        self.coefficients_pvalues = self._model.pvalues.copy()
        
        variables_included = [x for x in list(self.coefficients.index) if x!='Intercept']
        self._variables_included_patsy = variables_included.copy()
        
        variables_included = [LinearRegression._translate_from_patsy_notation(x) for x in variables_included]        
        
        self.variables_included = variables_included
        #self._independent_variables = 
        
        if self.include_constant:
            self._params_idx = ['Constant'] + variables_included
        else:
            self._params_idx = variables_included.copy()
        
        for stats in [self.coefficients,
                     self.coefficients_pvalues,
                     self.coefficients_sterrors,
                     self.coefficients_tvalues]:
            stats.index = self._params_idx    
        
        return
    
    @property
    def coefficients_beta(self):
        b = np.array(self._model.params)[1:]
        std_y = self._model.model.endog.std(axis=0)
        std_x = self._model.model.exog.std(axis=0)[1:]
        beta = list(b * (std_x / std_y))
        
        if self.include_constant:
            beta = [np.nan] + beta
            
        result = pd.Series(beta, index=self._params_idx)
        return result
    
    @property
    def coefficients_confidence_interval(self):
        
        ci = self._model.conf_int()
        ci.index = self._params_idx
        
        ci.columns = [f'LB CI (95%)',
                     f'UB CI (95%)']
        return ci
    
    @property
    def coefficients_VIF(self):
        #eps = 1e-20
        x = self._model.model.exog[:, 1:].copy()
        inv_corr = np.linalg.inv(np.corrcoef(x, rowvar=False))
        diag = list(inv_corr.diagonal())
        if self.include_constant:
            diag = [np.nan] + diag
        
        return pd.Series(diag, index=self._params_idx)
    
    @property
    def coefficients_tolerance(self):
        return 1 / self.coefficients_VIF
    
    @staticmethod
    def _translate_from_patsy_notation(effect):
        effect = effect\
        .replace(':', ' * ')\
        .replace('C(', '')\
        .replace('T.', '')\
        .replace('[', ' = "')\
        .replace(']', '"')\
        .replace(')', '')
        
        return effect
    
    def show_results(self, n_decimals):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int 
            Number of digits to round results when showing them
        """
        phrase = 'method {}'
        
        print('\nLINEAR REGRESSION SUMMARY')
        print('------------------\n')
        print('Model summary')
        display(self.summary_r2().style\
                    .set_caption(phrase.format('.summary_r2()'))\
                    .format(precision=n_decimals))
        print('------------------\n')
        print('ANOVA')
        display(self.summary_F().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(phrase.format('.summary_F()')))
        print('------------------\n')
        print('Coefficients')
        display(self.summary().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(phrase.format('.summary()')))
    
    def summary(self):
        """
        Summary table with requested information related to regression coefficients.

        Returns
        -------
        pd.DataFrame
            A summary table
        """
                  
        statistics = [
            self.coefficients,
            self.coefficients_sterrors,
            self.coefficients_beta,
            self.coefficients_tvalues,
            self.coefficients_pvalues
        ]
        
        columns = [
            'B', 
            'Std. Error',
            'Beta',
            't',
            'p-value'
        ]
        
        if self._show_ci:
            statistics.append(self.coefficients_confidence_interval)
            columns.extend(list(self.coefficients_confidence_interval.columns))
            
        if self._show_col:
            statistics.append(self.coefficients_tolerance)
            statistics.append(self.coefficients_VIF)
            columns.extend(['Tolerance', 'VIF'])
        
        statistics = pd.concat(statistics, axis=1)
        
        statistics.columns = columns
        
        statistics.index = self._params_idx
        
        return statistics
    
    def summary_r2(self):
        """
        Summary table with information related to coefficient of determination.

        Returns
        -------
        pd.DataFrame
            A summary table
        """
        r = self.r2 ** 0.5
        r2 = self.r2
        r2_adj = self.r2_adjusted
        
        statistics = [[r, r2, r2_adj]]
        columns = [
            'R',
            'R Squared',
            'Adj. R Squared'
        ]
        
        statistics = pd.DataFrame(
            statistics,
            columns=columns,
            index = ['']
        )
        
        return statistics
    
    def summary_F(self):
        """
        Summary table with information related to F-statistic.

        Returns
        -------
        pd.DataFrame
            A summary table
        """
        
        results = [[self.ess, self.dof_model, self.ms_model, self.F, self.F_pvalue],
                  [self.rss, self.dof_resid, self.ms_resid, np.nan, np.nan],
                  [self.tss, self.dof_total, np.nan, np.nan, np.nan]]
        
        results = pd.DataFrame(results,
                              columns = ['Sum of Squares', 'df', 'Mean Square', 'F', 'p-value'],
                              index = ['Regression', 'Residual', 'Total'])
                
        return results
    
    def _fit_backward(self):
        
        y_train = pd.Series(self._model.model.endog.copy(), 
                            name=self.dependent_variable,
                           index=self._observations_idx)
        X_train = pd.DataFrame(self._model.model.exog, columns=self._model.model.exog_names,
                              index=self._observations_idx)

        model = OLS(y_train, X_train, missing = 'drop')
        
        results = model.fit()
        
        max_pvalue = results.pvalues.drop('Intercept').max()
        
        while max_pvalue > self.sig_level_removal:
            x_to_drop = results.pvalues.drop('Intercept').idxmax()
            X_train = X_train.drop(x_to_drop, axis = 1)
            model = OLS(y_train, X_train, missing = 'drop')
            results = model.fit()
            max_pvalue = results.pvalues.drop('Intercept').max()
        
        self._model = results
        
        return
    
    def _identify_variables_without_variation(self):
        if self.include_constant:
            mask = self._model.model.exog.var(axis=0)[1:] == 0
        else:
            mask = self._model.model.exog.var(axis=0) == 0
            
        variables_included = [x for x in list(self._model.params.index) if x!='Intercept']

        return list(np.array(variables_included)[mask])
    
    def _remove_variables_without_variation(self):
        X = pd.DataFrame(self._model.model.exog, 
                         columns=self._model.model.exog_names, 
                        index=self._observations_idx)
        X = X.drop(self.variables_excluded, axis = 1)
        return X
                  
    def save_independent_variables(
        self,
        data=None,
        add_to_data=False
    ):
        """
        Produce values of independent variable remained in a fitted model.
        This option is useful if you don't create dummy variables or interaction effects manually
        but want to use them in a further analysis. Only variables remained in a model are returned
        (those that are shown in a summary table).
        
        Parameters
        ----------
        data : pd.DataFrame 
            Data for which independent variables are requested; 
            may be not specified if you want to save values for the same data that were used to fit a model
        add_to_data : bool 
            Whether to merge new values with the given data.
            Currently, this option returns data with a sorted index

        Returns
        -------
        pd.DataFrame
            Values of independent variables
        """
                  
        if data is None:
            data = self._data.copy()
            if self.include_constant:
                result = self._model.model.exog[:, 1:].copy()
            else:
                result = self._model.model.exog.copy()
            columns = [x for x in self.variables_included if x!='Constant']
            result = pd.DataFrame(
                result,
                columns=columns,
                index=self._observations_idx)
        
        else:
            aux_model = ols(self.formula, data).fit()
            aux_data_idx = aux_model.fittedvalues.index
            aux_data_cols = aux_model.model.exog_names
            aux_data_cols = [LinearRegression._translate_from_patsy_notation(x)\
                              for x in aux_data_cols]
            aux_data = pd.DataFrame(aux_model.model.exog,
                                    index=aux_data_idx,
                                    columns=aux_data_cols)
            result = aux_data[self.variables_included]
        
        if add_to_data:
            result = pd.concat([data, result], axis=1, sort=False)
                  
        return result
            
    def save_residuals(self,
                      unstandardized=True,
                      standardized=False,
                      studentized=False,
                      deleted=False,
                      studentized_deleted=False,
                      add_to_data=False):
        """
        Produce values of various residuals. 
        Residuals are returned only for data used to fit a model.
        
        Parameters
        ----------
        unstandardized : bool 
            Whether to save unstandardized (raw) residuals
        standardized : bool 
            Whether to save standardized (z-scores) residuals
        studentized : bool 
            Whether to save studentized residuals
        deleted : bool 
            Whether to save deleted residuals
        studentized_deleted : bool
            Whether to save studentized deleted residuals
        add_to_data : bool
            Whether to merge new values with data.
            Currently, this option returns data with a sorted index

        Returns
        -------
        pd.DataFrame
            Requested residuals
        """
                  
        columns_to_show = [f'{k.capitalize().replace("ized", ".").replace("eted", ".").replace("_", " ")} res.' \
                           for k, v in vars().items() if v==True and k!='add_to_data']
        
        infl = OLSInfluence(self._model)
        
        result = []
                  
        res_unstand = infl.resid
        res_unstand.name = 'Unstandard. res.'
                  
        res_stand = (res_unstand - res_unstand.mean()) / res_unstand.std()
        res_stand.name = 'Standard. res.'
                  
        res_stud = infl.resid_studentized_internal
        res_stud.name = 'Student. res.'
        
        result.extend([
            res_unstand, 
            res_stand,
            res_stud])
        
        if deleted:
            res_del = infl.resid_press
            res_del.name = 'Del. res.'
            result.append(res_del)                                                       
                           
        if studentized_deleted:
            res_stud_del = infl.resid_studentized_external
            res_stud_del.name = 'Student. del. res.'
            result.append(res_stud_del)      

        result = pd.concat(result, axis=1)
        result = result[columns_to_show].copy()
            
        if add_to_data:
            result = pd.concat([self._data, result], axis=1)
            
        return result
                           
    #following two methods are still in progress                       
    @staticmethod
    def _turn_all_rows_to_fancy(summary):
        return summary.apply(lambda x: LinearRegression._turn_one_row_to_fancy(x), axis=1)
    
    @staticmethod
    def _turn_one_row_to_fancy(row):
        coef = round(row['B'].item(), 3)
        sterr = round(row['Std. Error'].item(), 3)
        pval = row['p-value'].item()
        
        if pval <= 0.01:
            mark = '***'
        elif pval <= 0.05:
            mark = '**'
        elif pval <= 0.1:
            mark = '*'
        else:
            mark = ''
                           
        result = f'{coef}{mark} \n ({sterr})'
        return result

class BinaryLogisticRegression:
    
    """
    Class for binary logistic regression models based on the excellent statsmodels package.
    
    Parameters
    ----------
    method : 'enter' or 'backward' 
        Method for predictors selection 
    include_constant : bool 
        (CURRENTLY UNAVAILIABLE) Whether to include constant in the model
    classification_cutoff : float 
        Minimum probability to assign a prediction value 1
    sig_level_entry : float 
        (CURRENTLY UNAVAILIABLE) Max significance level to include predictor in the model 
    sig_level_removal : float 
        Min significance level to exclude predictor from the model

    Attributes
    ----------
    predictions : pd.Series
        Predicted values
    classification_table : pd.DataFrame
        A classification table
    precision_and_recall : pd.DataFrame
        Table with precision, recall, and F1-score of the model
    variables_excluded : list
        Variables excluded because of zero variance
    variables_included : list
        Variables included in a model
    N : int
        Number of observations included in a model
    r2_pseudo_macfadden : float
        MacFadden's pseudo coefficient of determination
    r2_pseudo_cox_snell : float
        Cox&Snell's pseudo coefficient of determination
    r2_pseudo_nagelkerke : float
        Nagelkerke's pseudo coefficient of determination
    loglikelihood : float
        -2LL
    coefficients : pd.Series
        Regression coefficients
    coefficients_sterrors : pd.Series
        Standard errors of regression coefficients
    coefficients_wald_statistics : pd.Series
        Wald statistic of regression coefficients
    coefficients_zvalues : pd.Series
        z-statistic of regression coefficients
    coefficients_pvalues : pd.Series
        P-values of regression coefficients
    coefficients_exp : pd.Series
        e ** regression coefficients
    """
    
    def __init__(
        self, 
        method='enter',
        include_constant=True,
        classification_cutoff=0.5,
        sig_level_entry=0.05,
        sig_level_removal=0.05,
    ):
        self.method = method.lower().strip()
        self.include_constant = include_constant
        self.classification_cutoff = classification_cutoff
        self.sig_level_entry = sig_level_entry
        self.sig_level_removal = sig_level_removal
    
    
    def fit(
        self,
        data,
        formula,
        categorical_variables=None,
        max_iterations=100,
        show_results=True,
        confidence_intervals=True,
        use_patsy_notation=False,
        n_decimals=3
    ):
        """
        Fit model to the given data using formula.

        Parameters
        ----------
        data : pd.DataFrame 
            Data to fit a model  
        formula : str 
            Formula of a model specification, e.g. 'y ~ x1 + x2'; 
            should be passed either in Patsy (statsmodels) notation
            or using the following rules: 
            '*' for interaction of the variables,
            ':' for interaction & main effects, 
            i.e., 'y ~ x:z' equals to 'y ~ x + z + x*z' (unlike the Patsy notation).
            If you use Patsy notation, please specify the parameter use_patsy_notation=True.
        categorical_variables : list 
            List of names of the variables that should be considered categorical.
            These variables would be automatically converted into sets of dummy variables.
            If you want to use this option, please make sure that you don't have nested names of variables
            (e.g. 'imdb' and 'imdb_rate' at the same time), otherwise this option results in an incorrect procedure.
        max_iterations : int 
            Maximum iterations for convergence
        show_results : bool 
            Whether to show results of analysis
        confidence_intervals : bool 
            Whether to include coefficients' confidence intervals in the summary table
        use_patsy_notation : bool 
            Turn this on if you use strictly Patsy's rules to define a formula.
            See more: https://patsy.readthedocs.io/en/latest/quickstart.html
        n_decimals : int 
            Number of digits to round results when showing them

        Returns
        -------
        self
            The current instance of the BinaryLogisticRegression class
        """
        
        self._data = data.copy()
        
        self.categorical_variables = categorical_variables
        self._show_ci = confidence_intervals
        self.max_iterations = max_iterations
        
        if '=' in formula:
            formula = formula.replace('=', '~')
        
        if not use_patsy_notation:
            formula = formula.replace('*', '^').replace(':', '*').replace('^', ':')
            
        self.formula = formula
        
        self.dependent_variable = self.formula.split('~')[0].strip()
        
        dep_cats = get_categories(self._data[self.dependent_variable])
        self._dep_cats = dep_cats
        
        if len(dep_cats) != 2:
            raise ValueError(f"""A dependent variable should have exactly 2 unique categories.
            The provided variable has {len(dep_cats)}.""")
            
        self._mapper = {dep_cats[0]: 0, dep_cats[1]: 1}
        self._inv_mapper = {0: dep_cats[0], 1: dep_cats[1]}
        
        if not is_numeric_dtype(self._data[self.dependent_variable]):
            self._data[self.dependent_variable] = self._data[self.dependent_variable].map(self._mapper).astype(int) 
        
        #won't work correctly if some variables have nested names (e.g. kinopoisk_rate and kinopoisk_rate_count)
        if categorical_variables is not None:
            if not isinstance(categorical_variables, list):
                raise ValueError(f"""Categorical variables should be passed as list.
                Type {type(categorical_variables)} was passed instead.""")
            else:
                for variable in categorical_variables:
                    formula = formula.replace(variable, f'C({variable})')
        self._optimizer = 'newton'
        try:
            self._model = logit(formula=formula, data=self._data).fit(
                maxiter=self.max_iterations, 
                warn_convergence=False,
                disp=False,
                method=self._optimizer,
                full_output=True
            )
        except np.linalg.LinAlgError:
            self._optimizer = 'bfgs'
            self._model = logit(formula=formula, data=self._data).fit(
                maxiter=self.max_iterations, 
                warn_convergence=False,
                disp=False,
                method=self._optimizer,
                full_output=True
            )
            
        self._model_params = {
            'maxiter': self.max_iterations,
            'warn_convergence': False,
            'disp': False,
            'method': self._optimizer,
            'full_output': True
        }
        
        self._observations_idx = list(self._model.fittedvalues.index)       
        self.variables_excluded = self._identify_variables_without_variation()
        
        if len(self.variables_excluded) > 0:
            y = pd.Series(self._model.model.endog.copy(), 
                          index=self._observations_idx,
                          name=self.dependent_variable)
            X = self._remove_variables_without_variation()
            self._model = Logit(y, X, missing = 'drop').fit(**self._model_params)
            self.variables_excluded = [BinaryLogisticRegression._translate_from_patsy_notation(x) for x in self.variables_excluded]
        
        
        if self.method == 'backward':
            self._fit_backward()
        
        self._get_statistics_from_model()
        
        self.predictions = self.predict()
        self.classification_table = self.get_classification_table()
        self.precision_and_recall = self.get_precision_and_recall()
        
        if show_results:
            self.show_results(n_decimals)
        
        if len(self.variables_excluded) > 0:
            print('------------------\n')
            print(f"Following variables were excluded due to zero variance: {'; '.join(self.variables_excluded)}")
        
        return self
    
    def _fit_backward(self):
        
        y_train = pd.Series(self._model.model.endog.copy(), 
                            name=self.dependent_variable,
                           index=self._observations_idx)
        X_train = pd.DataFrame(self._model.model.exog, columns=self._model.model.exog_names,
                              index=self._observations_idx)

        model = Logit(y_train, X_train, missing = 'drop')
        
        results = model.fit(**self._model_params)
        
        max_pvalue = results.pvalues.drop('Intercept').max()
        
        while max_pvalue > self.sig_level_removal:
            x_to_drop = results.pvalues.drop('Intercept').idxmax()
            X_train = X_train.drop(x_to_drop, axis = 1)
            model = Logit(y_train, X_train, missing = 'drop')
            results = model.fit(**self._model_params)
            max_pvalue = results.pvalues.drop('Intercept').max()
        
        self._model = results
        
        return
    
    def _identify_variables_without_variation(self):
        if self.include_constant:
            mask = self._model.model.exog.var(axis=0)[1:] == 0
        else:
            mask = self._model.model.exog.var(axis=0) == 0
            
        variables_included = [x for x in list(self._model.params.index) if x!='Intercept']

        return list(np.array(variables_included)[mask])
    
    def _remove_variables_without_variation(self):
        X = pd.DataFrame(self._model.model.exog, 
                         columns=self._model.model.exog_names, 
                        index=self._observations_idx)
        X = X.drop(self.variables_excluded, axis = 1)
        return X
    
    @staticmethod
    def _translate_from_patsy_notation(effect):
        effect = effect\
        .replace(':', ' * ')\
        .replace('C(', '')\
        .replace('T.', '')\
        .replace('[', ' = "')\
        .replace(']', '"')\
        .replace(')', '')
        
        return effect
                  
    def _get_statistics_from_model(self):
        
        self.N = self._model.nobs
        self.r2_pseudo_macfadden = self._model.prsquared
        self.r2_pseudo_cox_snell = 1 - np.exp(-self._model.llr/self.N)
        self.r2_pseudo_nagelkerke = self.r2_pseudo_cox_snell / (1 - np.exp(-(-2*self._model.llnull)/self.N))
        self.loglikelihood = -2 * self._model.llf
        
        self.coefficients = self._model.params.copy()
        self.coefficients_sterrors = self._model.bse.copy()
        self.coefficients_wald_statistics = self._model.tvalues.copy() ** 2
        self.coefficients_zvalues = self._model.tvalues.copy()
        self.coefficients_pvalues = self._model.pvalues.copy()
        self.coefficients_exp = self.coefficients.apply(np.exp)
        
        variables_included = [x for x in list(self.coefficients.index) if x!='Intercept']
        self._variables_included_patsy = variables_included.copy()
        
        variables_included = [BinaryLogisticRegression._translate_from_patsy_notation(x) for x in variables_included]        
        
        self.variables_included = variables_included
        
        if self.include_constant:
            self._params_idx = ['Constant'] + variables_included
        else:
            self._params_idx = variables_included.copy()
        
        for stats in [self.coefficients,
                     self.coefficients_pvalues,
                     self.coefficients_sterrors,
                     self.coefficients_zvalues,
                     self.coefficients_wald_statistics,
                     self.coefficients_exp]:
            stats.index = self._params_idx    
        
        return
                  
    def summary(self):
        """
        Summary table with requested information related to regression coefficients.
        
        Returns
        -------
        pd.DataFrame
            A summary table
        """
                  
        statistics = [
            self.coefficients,
            self.coefficients_sterrors,
            self.coefficients_wald_statistics,
            self.coefficients_pvalues,
            self.coefficients_exp
        ]
        
        columns = [
            'B', 
            'Std. Error',
            'Wald',
            'p-value',
            'Exp(B)'
        ]
        
        if self._show_ci:
            statistics.append(self.coefficients_confidence_interval)
            columns.extend(list(self.coefficients_confidence_interval.columns))
        
        statistics = pd.concat(statistics, axis=1)
        
        statistics.columns = columns
        
        statistics.index = self._params_idx
        
        return statistics
                  
    @property
    def coefficients_confidence_interval(self):
        
        ci = self._model.conf_int()
        ci.index = self._params_idx
        
        ci.columns = [f'LB CI (95%)',
                     f'UB CI (95%)']
        return ci
                  
    def show_results(self, n_decimals):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int 
            Number of digits to round results when showing them
        """
        phrase = 'method {}'
        
        print('\nLOGISTIC REGRESSION SUMMARY\n')
        if self._model.mle_retvals['converged']==True:
            print('Estimation was converged successfully.')
        else:
            print('Estimation was NOT converged successfully.')
            print('Please enlarge the number of iterations.')
        print('------------------\n')
        print('Dependent variable encoding')
        display(self.get_dependent_variable_codes().style\
                    .set_caption(phrase.format('.get_dependent_variable_codes()')))
        print('------------------\n')
        print('Model summary')
        display(self.summary_r2().style\
                    .set_caption(phrase.format('.summary_r2()'))\
                    .format(precision=n_decimals))
        print('------------------\n')
        print('Classification table')
        display(self.get_classification_table().style\
                    .set_caption(phrase.format('.get_classification_table()'))\
                    .format(precision=n_decimals))
        print('------------------\n')
        print('Precision and recall')
        display(self.get_precision_and_recall().style\
                    .set_caption(phrase.format('.get_precision_and_recall()'))\
                    .format(precision=n_decimals))
        print('------------------\n')
        print('Coefficients')
        display(self.summary().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(phrase.format('.summary()')))
        
    def summary_r2(self):
        """
        Summary table with information related to pseudo coefficients of determination.

        Returns
        -------
        pd.DataFrame
            A summary table
        """
        ll = self.loglikelihood
        mf = self.r2_pseudo_macfadden
        cs = self.r2_pseudo_cox_snell
        nk = self.r2_pseudo_nagelkerke
        
        statistics = [[ll, mf, cs, nk]]
        columns = [
            '-2 Log likelihood',
            "MacFadden's Pseudo R2",
            "Cox&Snell's Pseudo R2",
            "Nagelkerke's Pseudo R2",
        ]
        
        statistics = pd.DataFrame(
            statistics,
            columns=columns,
            index = ['']
        )
        
        return statistics
    
    def get_dependent_variable_codes(self):
        """
        Get information on how categories of a dependent variable were encoded.

        Returns
        -------
        pd.DataFrame
            A table explaining encodings
        """
        mapper = self._mapper
        result = pd.DataFrame(
            [list(mapper.items())[0], list(mapper.items())[1]],
            columns = ['Original value', 'Model value'],
            index = ['', ' ']
        )
        return result
    
    def get_classification_table(self):
        """
        Get a classification table.

        Returns
        -------
        pd.DataFrame
            A classification table
        """
        all_categories = self._dep_cats

        classification = pd.DataFrame(
            self._model.pred_table(),
            columns=self._dep_cats,
            index=self._dep_cats
        )

        classification.index.name = 'Observed'
        classification.columns.name = 'Predicted'
        classification['All'] = classification.sum(axis=1)
        classification.loc['All'] = classification.sum()

        n = classification.loc['All', 'All']
        for category in all_categories:
            classification.loc[category, 'All'] =  classification.loc[category, category] / classification.loc[category, 'All'] * 100
            classification.loc['All', category] =  classification.loc['All', category] / n * 100


        classification.loc['All', 'All'] = np.diagonal(classification.loc[all_categories, all_categories]).sum() / n * 100
        classification.index = all_categories + ['Percent predicted']
        classification.index.name = 'Observed'
        classification.columns = all_categories + ['Percent correct']
        classification.columns.name = 'Predicted'
        return classification
    
    def get_precision_and_recall(self):
        """
        Estimate precision, recall, and F-score for all the categories.

        Returns
        -------
        pd.DataFrame
            A table with estimated metrics
        """

        preds = self.classification_table.iloc[:-1, :-1]
        results = []
        categories = list(preds.index)
        for current_category in categories:
            idx = [cat for cat in categories if cat!=current_category]
            tp = preds.loc[current_category, current_category]
            fp = preds.loc[idx, current_category].sum()
            fn = preds.loc[current_category, idx].sum()
            if fp == 0:
                precision = 0
            else:
                precision = tp / (tp + fp)
            recall = tp / (tp + fn)
            if precision + recall != 0:
                f1 = 2 * (precision * recall) / (precision + recall)
            else:
                f1 = 0
            results.append([precision, recall, f1])
        results = pd.DataFrame(results, 
                               index=categories, 
                               columns = ['Precision', 'Recall', 'F score'])
        results.loc['Mean'] = results.mean()
        return results
    
    def predict(
        self,
        data=None,
        group_membership=True,
        probability=False,
        logit=False,
        add_to_data=False,
    ):
        """
        Predict values of a dependent variable using the fitted model.
        
        Parameters
        ----------
        data : pd.DataFrame 
            Data for prediction; 
            may be not specified if you want to predict values for the same data that were used to fit a model
        group_membership : bool 
            Whether to predict observation's membership 
            to categories of a dependent variable
        probability : bool 
            Whether to predict exact probability
        logit : bool 
            Whether to predict a logit value 
        add_to_data : bool 
            Whether to merge predictions with the given data.
            Currently, this option returns data with a sorted index

        Returns
        -------
        pd.DataFrame
            Predictions
        """
        name_memb = f'{self.dependent_variable} (predicted)'
        name_prob = f'{self.dependent_variable} (predicted prob.)'
        name_logit = f'{self.dependent_variable} (predicted logit)'
        
        all_columns = [name_memb, name_prob, name_logit]
        
        columns_to_show = []
        if group_membership:
            columns_to_show.append(name_memb)
        if probability:
            columns_to_show.append(name_prob)
        if logit:
            columns_to_show.append(name_logit)
        
        cutoff = self.classification_cutoff
        
        if data is None:
            data_init = self._data.copy()
            logit = self._model.fittedvalues
            prob = logit.apply(lambda x: np.exp(x) / (1 + np.exp(x)))
            memb = prob.apply(lambda x: 1 if x >= cutoff else 0).map(self._inv_mapper)
            result = pd.DataFrame(index=self._observations_idx, columns=all_columns)
            result[name_memb] = memb
            result[name_prob] = prob
            result[name_logit] = logit
            result = result[columns_to_show]
            if add_to_data:
                return pd.concat([data_init, result], axis=1)
            else:
                return result
                  
        else:
            aux_model = logit(self.formula, data).fit(**self._model_params)
            aux_data_idx = aux_model.fittedvalues.index
            aux_data_cols = aux_model.model.exog_names
            aux_data_cols = [BinaryLogisticRegression._translate_from_patsy_notation(x)\
                              for x in aux_data_cols]
            aux_data = pd.DataFrame(aux_model.model.exog,
                                    index=aux_data_idx,
                                    columns=aux_data_cols)
            aux_X = add_constant(aux_data[self.variables_included].copy())
            aux_y = aux_model.model.endog.copy()
                  
            aux_model = Logit(aux_y, aux_X, missing='drop').fit(**self._model_params)
            
            logit = aux_model.fittedvalues
            prob = logit.apply(lambda x: np.exp(x) / (1 + np.exp(x)))
            memb = prob.apply(lambda x: 1 if x >= cutoff else 0).map(self._inv_mapper)
            result = pd.DataFrame(index=aux_data_idx, columns=all_columns)
            result[name_memb] = memb
            result[name_prob] = prob
            result[name_logit] = logit
            result = result[columns_to_show]
            if add_to_data:
                return pd.concat([data, result], axis=1)
            else:
                return result
            
    def save_independent_variables(
        self,
        data=None,
        add_to_data=False
    ):
        """
        Produce values of independent variable remained in a fitted model.
        This option is useful if you don't create dummy variables or interaction effects manually
        but want to use them in a further analysis. Only variables remained in a model are returned
        (those that are shown in a summary table).
        
        Parameters
        ----------
        data : pd.DataFrame 
            Data for which independent variables are requested; 
            may be not specified if you want to save values for the same data that were used to fit a model
        add_to_data : bool 
            Whether to merge new values with the given data.
            Currently, this option returns data with a sorted index
        
        Returns
        -------
        pd.DataFrame
            Values of independent variables
        """
                  
        if data is None:
            data = self._data.copy()
            if self.include_constant:
                result = self._model.model.exog[:, 1:].copy()
            else:
                result = self._model.model.exog.copy()
            columns = [x for x in self.variables_included if x!='Constant']
            result = pd.DataFrame(
                result,
                columns=columns,
                index=self._observations_idx)
        
        else:
            aux_model = logit(self.formula, data).fit(**self._model_params)
            aux_data_idx = aux_model.fittedvalues.index
            aux_data_cols = aux_model.model.exog_names
            aux_data_cols = [BinaryLogisticRegression._translate_from_patsy_notation(x)\
                              for x in aux_data_cols]
            aux_data = pd.DataFrame(aux_model.model.exog,
                                    index=aux_data_idx,
                                    columns=aux_data_cols)
            result = aux_data[self.variables_included]
        
        if add_to_data:
            result = pd.concat([data, result], axis=1)
                  
        return result
    
    def save_residuals(self,
                      unstandardized=True,
                      standardized=False,
                      logit=False,
                      deviance=False,
                      add_to_data=False):
        """
        Produce values of various residuals. Residuals are returned only for data used to fit a model.
        
        Parameters
        ----------
        unstandardized : bool 
            Whether to save unstandardized (raw) residuals
        standardized : bool 
            Whether to save standardized (z-scores) residuals
        logit : bool 
            Whether to save logit residuals
        deviance : bool 
            Whether to save deviance residuals
        add_to_data : bool 
            Whether to merge new values with data.
            Currently, this option returns data with a sorted index

        Returns
        -------
        pd.DataFrame
            Requested residuals
        """
                  
        columns_to_show = [f'{k.capitalize().replace("ized", ".").replace("eted", ".").replace("_", " ")} res.' \
                           for k, v in vars().items() if v==True and k!='add_to_data']
        
        result = []
                  
        res_unstand = self._model.resid_response
        res_unstand.name = 'Unstandard. res.'
                  
        res_stand = self._model.resid_pearson
        res_stand.name = 'Standard. res.'
          
        res_deviance = self._model.resid_dev
        res_deviance.name = 'Deviance res.'                           
        
        preds_prob = self.predict(group_membership=False, probability=True)
        
        res_logit = res_unstand / (preds_prob * (1 - preds_prob)).iloc[:, 0]
        res_logit.name = 'Logit res.'                   
        
        result.extend([
            res_unstand, 
            res_stand,
            res_deviance,
            res_logit
        ])                                

        result = pd.concat(result, axis=1)
        
        result = result[columns_to_show].copy()
            
        if add_to_data:
            result = pd.concat([self._data, result], axis=1)
            
        return result



# coding: utf-8

'''
A module for formatting to a dataframe files of the formats: CSV, Excel and JSON. It facilitates working with data from social media
'''

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        # from google_drive_downloader import GoogleDriveDownloader
        import os, pandas
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '').replace('_', '')
        if '.' in module: module = module.split('.')[1]
        # if 'google_drive_downloader' in module: module = 'googledrivedownloader'
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break

def getFolder():
    folder = input('--- С какой директорией хотите работать? Укажите полный путь к ней, включая её саму:')
    folder = folder.replace('"', '') # поскольку в Windows путь из Проводника закавычен
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС
    return folder, slash

def getFolderFile(*arg): # арки: fileName и folder
    # print(*arg)
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС
    if len(arg) == 0:
        # print(0)
        folderFile = input('--- С каким файлом хотите работать? Укажите полный путь, включая название файла:')
        folderFile = folderFile.replace('"', '') # поскольку в Windows путь из Проводника закавычен
        # print(folderFile)
        fileName = folderFile.split(slash)[-1]
        # print(fileName)
        folder = slash.join(folderFile.split(slash)[:-1]) # из полного пути убрать имя файла
        folder += slash
        # print(folder)

    if len(arg) == 1:
        # print(1)
        fileName = arg[0]
        fileName = fileName.replace('"', '') # поскольку в Windows путь из Проводника закавычен
        # print(fileName)
        folder = ''
        # print(folder)

    if len(arg) == 2:
        # print(2)
        fileName = arg[0]
        fileName = fileName.replace('"', '') # поскольку в Windows путь из Проводника закавычен
        # print(fileName)
        folder = arg[1]
        folder = folder.replace('"', '') # поскольку в Windows путь из Проводника закавычен
        folder += slash
        # print(folder)
    return fileName, folder, slash

def excel2df(*arg):
    # print(*arg)
    fileName, folder, slash = getFolderFile(*arg)
    folderFile = folder + fileName
    # print(folderFile)
    error = None
    try:
        df = pandas.read_excel(folderFile, index_col=0)
        print(f'Файл "{folderFile}" импортирован')
    except:
        errorDescription = sys.exc_info()
        error = str(errorDescription[1]).replace("No module named '", '').replace("'", '').replace('_', '')
        print('error:', error)

    if error == None:
        display(df.head())
        print('Число столбцов:', df.shape[1], ', число строк', df.shape[0], '\n')
    else:
        df = None
    return df, error, fileName, folder, slash

def files2df(*arg):
    """
    Функция для оформления в датафрейм таблицы из файла формата Excel и присоединения к ней связанных ключом (id) таблиц из файлов форматов CSV, Excel и JSON, расположенных в той же директории

    Parameters
    ----------

    """
    # print(*arg)
    print('Эта функция предназначена для оформления в датафрейм таблицы из файла формата Excel'
          , 'и присоединения к ней связанных ключом (id) таблиц из файлов форматов CSV, Excel и JSON,'
          , 'расположенных в той же директории')
    df, error, fileName, folder, slash = excel2df(*arg)
    # print('fileName', fileName) # для отладки
    # print('folder', folder) # для отладки
    if error == None:
        if len(folder) == 0: # значит, из excel2df полный путь передан в fileName
            folder = slash.join(fileName.split(slash)[:-1]) + slash # из полного пути убрать имя файла
            fileName = fileName.split(slash)[-1] # из полного пути оставить только имя файла
    
        fileNameS = os.listdir(folder)
        print(fileNameS)
        fileNameS.remove(fileName)
    
        formatS = ['XLSX', 'CSV', 'JSON']
        for frmt in formatS:
            print('\n--- Если требуется найти ещё файл формата', frmt, 'для присоединения, то нажмите Enter'
                  , '\n--- Если НЕ требуется, то введите любой символ и нажмите Enter')
            if len(input()) == 0:
                fileNamesForImport = []
                for fileName in fileNameS:
                    if ('.' + frmt.lower()) in fileName:
                        print('--- Найден файл', fileName, '. Если он подходит, то нажмите Enter'
                              , '\n--- Если искать дальше, то введите любой символ и нажмите Enter')
                        if len(input()) == 0:
                            fileNamesForImport.append(fileName)
                            print('--- Файл', fileName, 'учтён; если требуется найти ещё файл формата', frmt, ', то нажмите Enter'
                                  , '\n--- Если НЕ требуется, то введите любой символ и нажмите Enter')
                            if len(input()) > 0:
                                break
    
                if len(fileNamesForImport) > 0:
                    # print('fileNamesForImport', fileNamesForImport) # для отладки
                    # print('folder', folder) # для отладки
                    # print('fileName', fileName) # для отладки
                    error = None
                    try:
                        for fileName in fileNamesForImport:
                            if 'xlsx' in fileName: df = df.merge(pandas.read_excel(f'{folder}{fileName}', index_col=0), on='id', how='outer')
                            if 'csv' in fileName: df = df.merge(pandas.read_csv(f'{folder}{fileName}'), on='id', how='outer')
                            if 'json' in fileName: df = df.merge(pandas.read_json(f'{folder}{fileName}'), on='id', how='outer')
                            # if 'json' in fileName: df = pandas.concat([df, pandas.read_json(f'{folder}{fileName}')], axis=1)
                    except:
                        errorDescription = sys.exc_info()
                        error = str(errorDescription[1])
                        print('error:', error)
                else:
                    print('Файлы форматов', formatS, 'не найдены в директории\n')
        if error == None:
            display(df.head())
            print('Число столбцов:', df.shape[1], ', число строк', df.shape[0], '\n')
        else:
            df = None
    return df, error, folder

# def googleDriver2local(file_id, dest_path):
#     GoogleDriveDownloader.download_file_from_google_drive(file_id=file_id, dest_path=dest_path)
#     error = None
#     try:
#         df = pandas.read_excel(dest_path, index_col=0)
#     except:
#         errorDescription = sys.exc_info()
#         error = str(errorDescription[1]).replace("No module named '", '').replace("'", '').replace('_', '')
#         print('error:', error)

#     if error == None:
#         display(df.head())
#         print('Число столбцов:', df.shape[1], ', число строк', df.shape[0], '\n')
#     else:
#         df = None
#     return df, error



#!/usr/bin/env python
# coding: utf-8

'''
A module for saving a dataframe to a file of one of the formats: CSV, Excel and JSON. It facilitates working with data from social media
'''

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        from randan.tools import varPreprocessor
        import os, pandas
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        if '.' in module: module = module.split('.')[1] 
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break

def df2file(dfIn, *arg): # арки: fileName и folder
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС

# ********** Выяснить поданные аргументы
    if len(arg) == 0:
        # print('len(arg) == 0')
        fileName = ''
        folder = ''

    if len(arg) == 1:
        # print('len(arg) == 1')
        fileName = arg[0]
        folder = ''

    if len(arg) == 2:
        # print('len(arg) == 2')
        fileName = arg[0]
        folder = arg[1]
        folder += slash

    if fileName == '':
        fileName = input('--- Впишите имя сохраняемого файла и нажмите Enter:')
    
    if folder != '':
        print(f'Директория, в которую сохраняю файл "{fileName}":', os.getcwd() + slash + folder)
    elif slash in fileName: # если директория содерджится в fileName
        folder = slash.join(fileName.split(slash)[:-1])
        fileName = fileName.split(slash)[-1]
        print(f'Директория, в которую сохраняю файл "{fileName}":', os.getcwd() + slash + folder)
    else:
        folder = input('--- Впишите директорию, в которую сохранить файл (если имя файла уже содержит путь к нему, то не вписывайте ничего) и нажмите Enter:')

    if slash != folder[-1]:
        folder += slash
# ********** Выяснить расширение сохраняемого файла
    # print('Имя сохраняемого файла:', fileName)
    fileFormatChoice = fileName.split('.')[-1]
    if (fileFormatChoice != 'xlsx') & (fileFormatChoice != 'csv') & (fileFormatChoice != 'json'):
        while True:
            print('--- Если хотите сохранить датафрейм в файл Excel, нажмите Enter;'
                  , '\n--- если же хотите в файл формата CSV или JSON, впишите букву "c" или "j" соответственно и нажмите Enter')
            fileFormatChoice = input()
            # print(folder + fileName.capitalize() + fileFormatChoice)
            if len(fileFormatChoice) == 0:
                fileFormatChoice = '.xlsx'
                break
            elif fileFormatChoice == 'c':
                fileFormatChoice = '.csv'
                break
            elif fileFormatChoice == 'j':
                fileFormatChoice = '.json'
                break
            else:
                print('--- Вы ввели что-то не то; попробуйте, пожалуйста, ещё раз..')
        fileName += fileFormatChoice

# ********** В зависимости от расширения сохраняемого файла выполнить сохранение
    # print('Расширение файла:', fileFormatChoice)
    # print('folder', folder) # для отладки
    # print('fileName', fileName) # для отладки
    if fileFormatChoice == 'xlsx':
        attempt = 0
        while True:
            try:
                dfIn.to_excel(folder + fileName)
                # print(folder + fileName)
                break
            except:
                errorDescription = sys.exc_info()
                print(errorDescription[1])
                if 'IllegalCharacterError' in str(errorDescription[0]):
                    module = 'xlsxwriter'
                    print('Для устранения ошибки требуется пакет', {module}, 'поэтому он будет инсталирован сейчас\n')
                    check_call([sys.executable, "-m", "pip", "install", module])
                    attempt += 1
                    if  attempt == 10:
                        print('Пакет', module
                              , 'НЕ прединсталлируется с установкой Анаконды, для работы скрипта требуется этот пакет,'
                              , 'но инсталлировать его не удаётся, попробуйте инсталлировать его вручную, после чего снова запустите требуемый скрипт пакета\n')
                        break
                else: break
    if fileFormatChoice == 'csv':
        dfIn.to_csv(folder + fileName)   
    if fileFormatChoice == 'json':
        dfIn.to_json(folder + fileName)

def df2fileShell(complicatedNamePart, dfIn, fileFormatChoice, method, coLabFolder, currentMoment):
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС
    folder = currentMoment + complicatedNamePart
    if coLabFolder == None:
        print('Сохраняю выгрузку метода', method, '                              ') #, f'в директорию "{folder}"'
        if os.path.exists(folder) == False:
            print('Такой директории не существовало, поэтому она создана')
            os.makedirs(folder)
        # else:
            # print('Эта директория существует')
    else:
        print('Сохраняю выгрузку метода', method, '                              ') #, f'в директорию "{os.getcwd() + slash + coLabFolder + slash + folder}"'
        if os.path.exists(os.getcwd() + slash + coLabFolder + slash + folder) == False:
            print('Такой директории не существовало, поэтому она создана')
            os.makedirs(os.getcwd() + slash + coLabFolder + slash + folder)
        # else:
            # print('Эта директория существует')
    
    # df2file(itemS) # при такой записи имя сохранаяемого файла и директория, в которую сохранить, вводятся вручную
    # print('При сохранении возможно появление обширного предупреждения UserWarning: Ignoring URL.'
    #       , 'Оно вызвано слишком длинными URL-адресами в датафрейме и не является проблемой; его следует пролистать и перейти к диалоговому окну' )

    # Проверка всех столбцов на наличие в их ячейках JSON-формата
    columnsToJSON = varPreprocessor.jsonChecker(dfIn)

    # print('folder', folder) # для отладки
    if len(columnsToJSON) > 0:
        print('В выгрузке метода', method, 'есть столбцы, содержащие внутри своих ячеек JSON-объекты; Excel не поддерживает JSON-формат;'
              , 'чтобы формат JSON не потерялся, сохраняю эти столбцы в файл формата НЕ XLSX, а JSON. Остальные же столбцы сохраняю в файл формата XLSX')

        columnsToJSON.append('id')
        if 'from_id' in dfIn.columns: columnsToJSON.append('from_id')
        if 'owner_id' in dfIn.columns: columnsToJSON.append('owner_id')

        df2file(dfIn[columnsToJSON], f'{folder}_{method}_JSON_varS.json', folder if coLabFolder == None else coLabFolder + slash + folder)
        columnsToJSON.remove('id')
        if 'from_id' in columnsToJSON: columnsToJSON.remove('from_id')
        if 'owner_id' in columnsToJSON: columnsToJSON.remove('owner_id')

        df2file(dfIn.drop(columnsToJSON, axis=1), f'{folder}_{method}_Other_varS{fileFormatChoice}', folder if coLabFolder == None else coLabFolder + slash + folder)
    else: df2file(dfIn, f'{folder}_{method}{fileFormatChoice}', folder if coLabFolder == None else coLabFolder + slash + folder)



import numpy as np
import pandas as pd
from scipy.stats import chi2, chi2_contingency, norm
from IPython.display import display
from scipy.stats import pearsonr, spearmanr, kendalltau
from pandas.api.types import is_numeric_dtype


class Crosstab:
    """
    
    Class to perform analysis of contingency tables based on Pearson's chi-square.
    
    Parameters
    ----------
    data : pd.DataFrame 
        Data to build a contingency table on
    row : str 
        Exact variable from the data to create rows of a contingency table
    column : str 
        Exact variable from the data to create columns of a contingency table
    sig_level : float
        Level of significance to analyze Pearson's residuals
    show_results : bool 
        Whether to show results of analysis
    n_decimals : int
        Number of digits to round results when showing them
    only_stats : bool 
        Use this if you only want to get final statistics,
        i.e. chi2, p-value, dof, and expected frequencies

    Attributes
    ----------
    N : int
        Number of observations used in the analysis
    n_cells : int
        Number of cells in a contingency table 
    frequencies_observed : pd.DataFrame
        Observed frequencies
    frequencies_expected : pd.DataFrame
        Expected frequencies
    residuals : pd.DataFrame
        'Raw' residuals 
        (differences between obs. and exp. frequencies)
    residuals_pearson : pd.DataFrame
        Pearson's residuals
        (raw residuals divided by square root of expected frequencies)
    chi_square: float
        Chi-square statistics
    dof : int
        Degrees of freedom
    pvalue : float
        P-value of a chi-square statistic
    row_categories : list
        Names of categories located at rows of a contingency table
    column_categories : list
        Names of categories located at columns of a contingency table
    row_marginals : pd.DataFrame
        Marginal frequencies of row categories
    column_marginals : pd.DataFrame
        Marginal frequencies of column categories
    """        
    
    def __init__(self, 
                 data, 
                 row=None, 
                 column=None,
                 sig_level=0.05,
                 show_results=True,
                 n_decimals=3,
                 only_stats=False):
              
        if row is not None and column is not None:
            self.data = data[[row, column]]
        elif (row is None or column is None) and len(data.columns) > 2:
            raise ValueError('''Number of variables in dataframe exceeds 2.
            Please, specify both row and column arguments or filter the necessary variables manually.''')
        elif len(data.columns) < 2:
            raise ValueError('''One or no variables were passed.''')
        else:
            self.data = data
        
        self.sig_level = sig_level
        
        self._frequencies_observed = pd.crosstab(self.data.iloc[:, 0],
                                                self.data.iloc[:, 1])
        
        if only_stats:
            self.chi_square, self.pvalue, self.dof, self.frequencies_expected = \
            chi2_contingency(self._frequencies_observed)
            return
        
        self.N = self._frequencies_observed.sum().sum()
        
        self.n_cells = self._frequencies_observed.shape[0] * self._frequencies_observed.shape[1]        
        
        self.row_categories = list(self._frequencies_observed.index)
        self.column_categories = list(self._frequencies_observed.columns)
        
        self.frequencies_observed = self._frequencies_observed.copy()
        
        self.row_marginals = Crosstab._get_marginals(self._frequencies_observed, 'row')       
        self.column_marginals = Crosstab._get_marginals(self._frequencies_observed, 'column')
        
        self.frequencies_observed = Crosstab._add_marginals(self._frequencies_observed,
                                                            self.row_marginals,
                                                            self.column_marginals,
                                                            self.N)

        self._frequencies_expected = self._get_expected_frequencies()
               
        self.frequencies_expected = Crosstab._add_marginals(self._frequencies_expected,
                                                            self.row_marginals,
                                                            self.column_marginals,
                                                            self.N)
        self.residuals = self._frequencies_observed - self._frequencies_expected
        self.residuals_pearson = self.residuals / (self._frequencies_expected**0.5)
        self.chi_square = (self.residuals_pearson**2).sum().sum()
        self.dof = (len(self.row_categories) - 1)*(len(self.column_categories) - 1)
               

        self.pvalue = chi2.sf(self.chi_square, self.dof)
        
        if show_results:
            self.show_results(n_decimals)
    
    @staticmethod
    def _get_marginals(crosstab, axis='row'):
        if axis=='row':            
            marginals = pd.DataFrame(crosstab.sum(axis=1),
                                     columns=['Total'])        
        elif axis=='column':
            marginals = pd.DataFrame(crosstab.sum(),
                                     columns=['Total']).T
        else:
            raise ValueError('Unknown axis.')
        
        return marginals    
    
    @staticmethod
    def _add_marginals(crosstab, row_marginals, column_marginals, N=None):
        if N is None:
            N = crosstab.sum().sum()
        
        crosstab = crosstab.copy()
        
        #weird but the only way that works though        
        try:
            crosstab.columns = crosstab.columns.add_categories('Total')
            crosstab['Total'] = row_marginals
        except AttributeError:
            crosstab['Total'] = row_marginals
        
        try:    
            crosstab.index = crosstab.index.add_categories('Total')
            crosstab.loc['Total'] = column_marginals.T['Total'].tolist() + [N]
        except AttributeError:
            crosstab.loc['Total'] = column_marginals.T['Total'].tolist() + [N]        
        
        return crosstab
    
    def _get_expected_frequencies(self):
        frequencies_expected = np.array([self.column_marginals.T['Total'].to_list()] * len(self.row_categories))
        frequencies_expected = frequencies_expected * self.row_marginals['Total'].to_numpy()[:, np.newaxis] / self.N
        frequencies_expected = pd.DataFrame(frequencies_expected,
                                            columns=self._frequencies_observed.columns,
                                            index=self._frequencies_observed.index)
        return frequencies_expected
        
    def check_small_counts(self, n_decimals=3):
        """
        Check sparsity of the crosstab, i.e., 
        calculate percentage of small expected frequencies
        (those that are less than 5).

        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them

        """
        
        n_small_counts = (self._frequencies_expected < 5).sum().sum()
        min_count = self._frequencies_expected.min().min()
        n_cells = self.n_cells
        perc_small_counts = n_small_counts / n_cells * 100
        print(f'''{n_small_counts} ({round(perc_small_counts, n_decimals)}%) cells have expected \
frequency less than 5. The minimum expected frequency is {round(min_count, n_decimals)}.''')
        
    def check_significant_residuals(self, sig_level=0.05, n_decimals=3):
        """
        Identify significant Pearson's residuals 
        based on the given significant level.

        Parameters
        ----------
        sig_level : float
            Significance level (alpha) to identify significant residuals
        n_decimals : int
            Number of digits to round results when showing them
        """
        
        critical_value = norm.isf(sig_level / 2)
        n_sig_resids = (abs(self.residuals_pearson) >= critical_value).sum().sum()
        n_cells = self.n_cells
        perc_sig_resids = n_sig_resids / n_cells * 100
        
        max_resid = self.residuals_pearson.max().max()
        max_resid_row = self.residuals_pearson.max(axis=1).idxmax()
        max_resid_column = self.residuals_pearson.max(axis=0).idxmax()
        
        min_resid = self.residuals_pearson.min().min()
        min_resid_row = self.residuals_pearson.min(axis=1).idxmin()
        min_resid_column = self.residuals_pearson.min(axis=0).idxmin()
        print(f'''{n_sig_resids} ({round(perc_sig_resids, n_decimals)}%) cells have Pearson's \
residual bigger than {round(critical_value, 2)}. 
The biggest residual is {round(max_resid, n_decimals)} (categories {max_resid_row} and {max_resid_column}).
The smallest residual is {round(min_resid, n_decimals)} (categories {min_resid_row} and {min_resid_column}).''')
        
    def show_results(self, n_decimals=3):
        """
        Show results of the analysis in a readable form.

        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them
        """
        
        sig_level = self.sig_level
        
        print('\nCROSSTAB SUMMARY')
        print('------------------\n')
        print('Observed frequencies')
        display(self.frequencies_observed.style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("attribute .frequencies_observed"))
        print('------------------\n')
        print('Expected frequencies')
        display(self.frequencies_expected.style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("attribute .frequencies_expected"))
        self.check_small_counts(n_decimals)
        print('------------------\n')
        print(f'Chi-square statistic is {round(self.chi_square, n_decimals)} (p-value = {round(self.pvalue, n_decimals)}).')
        print('------------------\n')
        print("Pearson's residuals")
        display(self.residuals_pearson.style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("attribute .residuals_pearson"))
        self.check_significant_residuals(sig_level, n_decimals)

#todo: flag significant, pairwise deletion

class Correlation:
    """
    
    Class to perform correlation analysis.
    
    Parameters
    ----------
    data : pd.DataFrame 
        Data to perform the analysis
    variables : list
        Variables from data to include in the analysis
    method : str
        Which correlation coefficient to use.
        Available values: 'pearson', 'spearman', 'kendall'
    two_tailed : bool
        Whether to calculate two-tailed p-value (set to False if you are interested in one-tailed p-value)
    show_results : bool 
        Whether to show results of analysis
    n_decimals : int
        Number of digits to round results when showing them
    min_max : bool
        Whether to print minimum and maximum correlation coefficients
    Attributes
    ----------
    tailes : int
        Number of tailes (1 or 2) to estimate p-value
    method : int
        Type of correlation coefficient  
    N : int
        Number of observations included in the analysis
    correlation_matrix : pd.DataFrame
        Correlation matrix
    """ 
    def __init__(
        self,
        data,
        variables=None,
        method='pearson',
        two_tailed=True,
        show_results=True,
        n_decimals=3, 
        min_max=True
    ):
        if variables is not None and not isinstance(variables, list):
            raise TypeError(f'Variables should be passed as list. Type {type(variables)} was passed instead.')
            
        if not isinstance(method, str):
            raise TypeError(f'Method should be passed as str. Type {type(method)} was passed instead.')
            
        if variables is None:
            variables = data.columns
        
        non_num_vars = [var for var in variables if not is_numeric_dtype(data[var])]
        zero_var_vars = [var for var in variables if is_numeric_dtype(data[var]) and data[var].var()==0]
        variables = [var for var in variables if var not in non_num_vars and var not in zero_var_vars]
        
        self._non_num_vars = []
        self._zero_var_vars = []
        
        if len(non_num_vars) > 0:
            self._non_num_vars = non_num_vars
        if len(zero_var_vars) > 0:
            self._zero_var_vars = zero_var_vars    
        
        if len(variables) < 2:
            raise ValueError('One or zero numeric variables were passed or found in a dataframe.')
        
        data = data[variables].dropna().copy()
        
        self.tailes = 2 if two_tailed else 1
        
        if method.lower() == 'pearson':
            metric = pearsonr
        elif method.lower() == 'spearman':
            metric = spearmanr
        elif method.lower() == 'kendall':
            metric = kendalltau
#         elif method.lower() == 'all':
#             pass
        else:
            raise ValueError("Unknown method requested. Possible values: 'pearson', 'spearman', 'kendall'")
            
        self.method = method
        
        N = len(data)
        self.N = N
        
        stats = ['Coefficient', 'p-value', 'N']
        index = [variables, stats]
        index = pd.MultiIndex.from_product(index)
        results = pd.DataFrame(index=index)
        
        corr_dict = {}
        
        for var_1 in variables:
            for var_2 in variables:
                if var_1 == var_2:
                    r, pval = 1, np.nan
                else:
                    r, pval = metric(data[var_1], data[var_2])
                    if not two_tailed:
                        pval /= 2
                    if (var_2, var_1) not in corr_dict.keys():
                        corr_dict.update({(var_1, var_2): (r, pval)})
                
                results.loc[(var_1, 'Coefficient'), var_2] = r
                results.loc[(var_1, 'p-value'), var_2] = pval
                results.loc[(var_1, 'N'), var_2] = N
        
        self._corr_dict = corr_dict
        
        self.correlation_matrix = results
        
        if show_results:
            self.show_results(n_decimals, min_max)
            
    def show_results(self, n_decimals=3, min_max=True):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them
        min_max : bool
            Whether to print minimum and maximum correlation coefficients
        """
        print(f'\nCORRELATION SUMMARY ({self.method.upper()} METHOD, {self.tailes}-TAILED)')
        print('------------------')
        
        if len(self._non_num_vars) > 0:
            phrase = 'The following variables were removed from the analysis since they do not belong to numerical dtypes: {}\n'
            print(phrase.format(', '.join(self._non_num_vars)))
        if len(self._zero_var_vars) > 0:
            phrase = 'The following variables were removed from the analysis since they have zero variance: {}\n'
            print(phrase.format(', '.join(self._zero_var_vars)))
        display(self.correlation_matrix.style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("attribute .correlation_matrix"))
        if min_max:
            results = self.sort_correlations()
            self.min_corr = results['Coefficient'].min()
            idxmin = results['Coefficient'].idxmin()
            min_pval = round(results.loc[idxmin, 'p-value'], n_decimals)
            self.max_corr = results['Coefficient'].max()
            idxmax = results['Coefficient'].idxmax()
            max_pval = round(results.loc[idxmax, 'p-value'], n_decimals)
            print(f'Maximum correlation is {round(self.max_corr, n_decimals)} (p-value {max_pval}) for variables {idxmax[0]} and {idxmax[1]},',
                  f'minimum correlation is {round(self.min_corr, n_decimals)} (p-value {min_pval}) for variables {idxmin[0]} and {idxmin[1]}.', 
                  sep='\n')
            
    def sort_correlations(self, by='coefficient', ascending=True):
        """
        Sort correlations between all possible pairs of variables.

        Parameters
        ----------
        by : str
            Which parameter should define sorting: 'coefficient' or 'pvalue'
        ascending : bool
            Whether to sort values in ascending order
        """
        coefs = pd.DataFrame(self._corr_dict, index=[0, 1]).T
        coefs.columns = ['Coefficient', 'p-value']
        if by.lower() == 'coefficient':
            by = 'Coefficient'
        elif by.lower() in ['p-value', 'pvalue', 'p_value']:
            by = 'p-value'
        print('Note: Each empty index duplicates the previous one.')
        return coefs.sort_values(by=by, ascending=ascending)



from itertools import combinations
from scipy.stats import norm
import pandas as pd
import numpy as np

#extend this list
available_colors = [
    'cadetblue1',
    'bisque',
    'pink',
    'aquamarine',
    'olivedrab1',
    'slategray1',
    'darkseagreen1'
]

def binning(series, n_intervals):
    series = pd.qcut(series, n_intervals, duplicates='drop')
    return series

def merge_two_cats(series, cat1, cat2):
    joint_cat = str(cat1) + ' / ' + str(cat2)
    return series.apply(lambda x: joint_cat if x in (cat1, cat2) else x).astype(str)

def merge_two_intervals(series, cat1, cat2):
    #display(series, cat1, cat2)
    pair = cat1, cat2
    lower_bound = cat1.left if cat1.left < cat2.left else cat2.left
    upper_bound = cat2.right if cat2.right > cat1.right else cat1.right
    return series.apply(lambda x: pd.Interval(lower_bound, upper_bound) if x in pair else x)

def get_categories(series):
    return sorted(series.dropna().unique())

def get_unordered_combinations(categories, k):
    return list(combinations(categories, k))

def get_ordered_combinations(categories, k):
    categories_combo = [combo for combo in combinations(categories, k)]
    categories_combo = [combo for combo in categories_combo \
                        if abs(categories.index(combo[0])-categories.index(combo[1]))==1]
    return categories_combo

#move to descriptive statistics
def confidence_interval_mean(mean, std, n, sig_level=0.05):
    z_crit = norm.isf(sig_level / 2)
    error = z_crit * (std/(n**(1/2)))
    if pd.isnull(error):
        error = 0
    lower_bound = mean - error
    upper_bound = mean + error
    return pd.Interval(lower_bound, upper_bound, closed='both')

#move to descriptive statistics ?
def confidence_interval_comparison(first_ci, second_ci):
    if first_ci.overlaps(second_ci):
        return 'overlap'
    elif first_ci < second_ci:
        return 'lower'
    elif first_ci > second_ci:
        return 'bigger'

def classification_table(y_true, y_predicted):
    classification = pd.crosstab(y_true, 
                                    y_predicted,
                                    margins=True)
    classification.index.name = 'Observed'
    classification.columns.name = 'Predicted'
    all_categories = list(classification.index)
    empty_categories = [value for value in all_categories if value not in classification.columns]
    for category in empty_categories:
        classification[category] = 0
    classification = classification[all_categories]

    all_categories.remove('All')
    
    n = classification.loc['All', 'All']
    
    for category in all_categories:
        classification.loc[category, 'All'] =  classification.loc[category, category] / classification.loc[category, 'All'] * 100
        classification.loc['All', category] =  classification.loc['All', category] / n * 100
    
    
    classification.loc['All', 'All'] = np.diagonal(classification.loc[all_categories, all_categories]).sum() / n * 100
    classification.index = all_categories + ['Percent predicted']
    classification.index.name = 'Observed'
    classification.columns = all_categories + ['Percent correct']
    classification.columns.name = 'Predicted'
    return classification

def precision_and_recall(classification_table):
    """
    Estimate precision, recall, and F-score for all the categories.
    """

    preds = classification_table.iloc[:-1, :-1]
    results = []
    categories = list(preds.index)
    for current_category in categories:
        idx = [cat for cat in categories if cat!=current_category]
        tp = preds.loc[current_category, current_category]
        fp = preds.loc[idx, current_category].sum()
        fn = preds.loc[current_category, idx].sum()
        if fp == 0:
            precision = 0
        else:
            precision = tp / (tp + fp)
        recall = tp / (tp + fn)
        if precision + recall != 0:
            f1 = 2 * (precision * recall) / (precision + recall)
        else:
            f1 = 0
        results.append([precision, recall, f1])
    results = pd.DataFrame(results, 
                            index=categories, 
                            columns = ['Precision', 'Recall', 'F score'])
    results.loc['Mean'] = results.mean()
    return results



#!/usr/bin/env python
# coding: utf-8

'''
(EN) A module for processing two (semi)standardized text variables: brandingSettings.channel.keywords or snippet.tags and topic Details.topic Categories. They are variables from the output of the YouTube API methods `channels` or `videos`. As well the module creates a column with a channels's or video's absolute link
(RU) Авторский модуль для обработки двух (полу)стандартизированных текстовых переменных выгрузки методов channels или videos API YouTube: brandingSettings.channel.keywords или snippet.tags и topicDetails.topicCategories, а также создания столбца с абсолютной ссылкой на канала или видео
'''

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys, warnings
from randan.tools import textPreprocessor
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        import pandas, warnings
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '').replace('_', '')
        if '.' in module: module = module.split('.')[1]
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break

def keywordTagTopicURL(dfIn):
    """
    Функция для обработки двух (полу)стандартизированных текстовых переменных выгрузки метода videos: snippet.tags, topicDetails.topicCategories, а также создания столбца с абсолютной ссылкой

    Parameters
    ----------
    dfIn : DataFrame
    """
    df = dfIn.copy()
    contentType = 'channel' if 'brandingSettings.channel.keywords' in df.columns else 'video'
    
    warnings.filterwarnings("ignore")
    varS = ['brandingSettings.channel.keywords' if contentType == 'channel' else 'snippet.tags', 'topicDetails.topicCategories']
    for var in varS:
        nanDf = df[df[var].isna()]
        notNanDf = df[df[var].notna()]
        if var == 'brandingSettings.channel.keywords': notNanDf[var] = notNanDf[var].str.split(' ')
        notNanDf[var] = notNanDf[var].apply(lambda cellContent: ' '.join(cellContent))
        if var == 'topicDetails.topicCategories': notNanDf['https://en.wikipedia.org/wiki/'] = notNanDf[var].str.replace('https://en.wikipedia.org/wiki/', '')
        else: # расщепить на отдельные слова и удалить среди них дубликаты
            # print('var:', var) # для отладки
            notNanDf[var] = notNanDf[var].apply(lambda cellContent: textPreprocessor.simbolsCleaner(cellContent))
            notNanDf[var] = textPreprocessor.pymystemLemmatizer(notNanDf, var)
            notNanDf[var] = notNanDf[var].apply(lambda cellContent: textPreprocessor.stopwordsDropper(cellContent))
        df = pandas.concat([nanDf, notNanDf])
    
    df['URL'] = f'https://www.youtube.com/{contentType}/' + df['id'] # столбец с абсолютной ссылкой
    
    display(df[[varS[0], 'https://en.wikipedia.org/wiki/', 'URL']].tail())
    print('Число строк', df.shape[0])
    return df



#!/usr/bin/env python
# coding: utf-8


# In[ ]:


# # 0 Активировать требуемые для работы скрипта модули и пакеты + пререквизиты


# In[ ]:


# 0.0 В общем случае требуются следующие модули и пакеты (запасной код, т.к. они прописаны в setup)
# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        from datetime import date, datetime
        from randan.tools import calendarWithinYear # авторский модуль для работы с календарём конкретного года
        from randan.tools import coLabAdaptor # авторский модуль для адаптации текущего скрипта к файловой системе CoLab
        from randan.tools import df2file # авторский модуль для сохранения датафрейма в файл одного из форматов: CSV, Excel и JSON в рамках работы с данными из социальных медиа
        from randan.tools import files2df # авторский модуль для оформления в датафрейм таблиц из файлов формата CSV, Excel и JSON в рамках работы с данными из социальных медиа
        from randan.tools import varPreprocessor # авторский модуль для предобработки переменных номинального, порядкового, интервального и более высокого типа шкалы
        import numpy, os, pandas, re, shutil, time, requests, warnings
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        if '.' in module: module = module.split('.')[0]
        print(
f'''Пакет {module} НЕ прединсталлирован, но он требуется для работы скрипта, поэтому будет инсталлирован сейчас
Попытка № {attempt} из 10
'''
              )
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print(
f'''Пакет {module} НЕ прединсталлирован; он требуется для работы скрипта, но инсталлировать его не удаётся,
поэтому попробуйте инсталлировать его вручную, после чего снова запустите скрипт
'''
                  )
            break


# In[ ]:


# # 1 Авторские функции


# In[ ]:


# 1.0 для метода search из API ВК, помогающая работе с ключами
def bigSearch(
              API_keyS,
              count,
              end_time,
              fields,
              iteration,
              keyOrder,
              latitude,
              longitude,
              pause,
              q,
              start_from,
              start_time
              ):
    # print('    start_from', start_from) # для отладки
    dfAdd = pandas.DataFrame()
    goS = True
    params = {
              'access_token': API_keyS[keyOrder], # обязательный параметр
              'count': count, # опциональный параметр
              'end_time': end_time, # опциональный параметр
              'extended': 1, # опциональный параметр
              'fields': fields, # опциональный параметр
              'latitude': latitude, # опциональный параметр
              'longitude': longitude, # опциональный параметр
              'q': q, # опциональный параметр
              'start_from': start_from, # опциональный параметр
              'start_time': start_time, # опциональный параметр
              'v': '5.199' # обязательный параметр
              }
    while True:
        response = requests.get('https://api.vk.ru/method/newsfeed.search', params=params)
        response = response.json() # отобразить выдачу метода get в виде JSON
        # print('response', response) # для отладки
        if 'response' in response.keys():
            response = response['response']
            # print('    response.keys() внутри bigSearch', response.keys()) # для отладки
            dfAdd = pandas.json_normalize(response['items'])
            break
        elif 'error' in response.keys():
            if 'Too many requests per second' in response['error']['error_msg']:
                # print('  keyOrder до замены', '                    ') # для отладки
                keyOrder = keyOrder + 1 if keyOrder < (len(API_keyS) - 1) else 0 # смена ключа, если есть на что менять
                print(f'\nПохоже, ключ попал под ограничение вследствие слишком высокой частоты обращения скрипта к API; пробую перейти к следующему ключу (№ {keyOrder}) и снизить частоту')
                # print('  keyOrder после замены', keyOrder, '                    ') # для отладки
                pause += 0.25

            elif 'Unknown application: could not get application' in response['error']['error_msg']:
                # print('  keyOrder до замены', '                    ') # для отладки
                keyOrder = keyOrder + 1 if keyOrder < (len(API_keyS) - 1) else 0 # смена ключа, если есть на что менять
                print('\nПохоже, Ваше ВК-приложение попало под ограничение; пробую перейти к следующему ключу (№ {keyOrder}) и снизить частоту')
                # print('  keyOrder после замены', keyOrder, '                    ') # для отладки
                pause += 0.25

            elif 'Internal server error: Unknown error, try later' in response['error']['error_msg']:
                print('\nПохоже, ошибка на сервере ВК; подождите и запустите скрипт с начала')
                response = {'items': [], 'total_count': 0} # принудительная выдача для response
                goS = False # нет смысла продолжать исполнение скрипта
                break # и, следовательно, нет смысла в новых итерациях цикла

            elif 'User authorization failed' in response['error']['error_msg']:
                print(
'''
Похоже, аккаунт попал под ограничение. Оно может быть снято с аккаунта сразу или спустя какое-то время. Подождите или подготовьте новый ключ в другом аккаунте. И запустите скрипт с начала'''
                      )
                response = {'items': [], 'total_count': 0} # принудительная выдача для response
                goS = False # нет смысла продолжать исполнение скрипта
                break # и, следовательно, нет смысла в новых итерациях цикла

            else:
                print('  Похоже, проблема НЕ в слишком высокой частоте обращения скрипта к API((')
                print('  ', response['error']['error_msg'])
                goS = False # нет смысла продолжать исполнение скрипта
                break # и, следовательно, нет смысла в новых итерациях цикла

    # Для визуализации процесса
    print('    Итерация №', iteration, ', number of items', len(response['items']), '                    ', end='\r')
    iteration += 1

    # Сменить формат представления дат, класс данных столбцов с id, создать столбец с кликабельными ссылками на контент
        # Здесь, а не в конце, поскольку нужна совместимость с itemS из Temporal и от пользователя
    if len(dfAdd) > 0:
        # display(dfAdd) # для отладки
        # print('dfAdd.columns', dfAdd.columns) # для отладки
        dfAdd['date'] = dfAdd['date'].apply(lambda content: datetime.fromtimestamp(content).strftime('%Y.%m.%d'))
        dfAdd['URL'] = dfAdd['from_id'].astype(str)
        dfAdd.loc[dfAdd[dfAdd['URL'].str.contains('-') == False].index, 'URL'] = 'id' + dfAdd.loc[dfAdd[dfAdd['URL'].str.contains('-') == False].index, 'URL']
        dfAdd.loc[dfAdd[dfAdd['URL'].str.contains('-')].index, 'URL'] = dfAdd.loc[dfAdd[dfAdd['URL'].str.contains('-')].index, 'URL'].str.replace('-', 'public')
        dfAdd['URL'] = 'https://vk.com' + '/' + dfAdd['URL'] + '?w=' + dfAdd['inner_type'].str.split('_').str[0] + dfAdd['owner_id'].astype(str) + '_' + dfAdd['id'].astype(str)

        if fields != None:
            for fieldsColumn in ['groups', 'profiles']:
                if fieldsColumn in response.keys(): dfAdd = fieldsProcessor(dfIn=dfAdd, fieldsColumn=fieldsColumn, response=response)

    return dfAdd, goS, iteration, keyOrder, pause, response

# 1.1 для обработки выдачи любого из методов, помогающая работе с ключами
def dfsProcessor(
                 complicatedNamePart,
                 coLabFolder,
                 dfAdd,
                 dfFinal, # на обработке какой бы ни было выгрузки не возникла бы непреодолимая ошибка, сохранить следует выгрузку метода search
                 dfIn,
                 fileFormatChoice,
                 goS, # единственная из функций, принимающая этот аргумент
                 method,
                 momentCurrent,
                 q,
                 slash,
                 stage,
                 targetCount,
                 year,
                 yearsRange
                 ):
    df = pandas.concat([dfIn, dfAdd])
    columnsForCheck = []
    if columnsForCheck == []: # для выдач, НЕ содержащих столбец id, проверка дублирующихся  строк возможна по столбцам, содержащим в имени id
        for column in df.columns:
            if 'id' in column: columnsForCheck.append(column)
    # print('Столбцы, по которым проверяю дублирующиеся строки:', columnsForCheck) # для отладки
    df = df.drop_duplicates(columnsForCheck, keep='last').reset_index(drop=True) # при дублировании объектов из itemS из Temporal и от пользователя и новых объектов, оставить новые

    if goS == False:
        print(
f'Поскольку исполнение скрипта натолкнулось на ошибку, сохраняю выгруженный контент и текущий этап поиска в директорию "{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal"'
              )
        if not os.path.exists(f'{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal'):
                os.makedirs(f'{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal')
                print(f'Директория "{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal" создана')
        # else:
            # print(f'Директория "{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal" существует')

# Сохранение следа исполнения скрипта, натолкнувшегося на ошибку, непосредственно в директорию Temporal в текущей директории
        file = open(f'{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal{slash}method.txt', 'w+') # открыть на запись
        file.write(method)
        file.close()

        file = open(f'{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal{slash}q.txt', 'w+') # открыть на запись
        file.write(q if q != None else '')
        file.close()
    
        file = open(f'{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal{slash}stageTarget.txt', 'w+')
        file.write(str(stageTarget)) # stageTarget принимает значения [0; 3]
        file.close()
    
        file = open(f'{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal{slash}targetCount.txt', 'w+')
        file.write(str(targetCount))
        file.close()
    
        file = open(f'{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal{slash}year.txt', 'w+')
        file.write(str(year)) # год, на котором остановилось исполнение скрипта
        file.close()

        file = open(f'{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal{slash}yearsRange.txt', 'w+')
        file.write(yearsRange if yearsRange != None else '') # пользовательский временнОй диапазон
        file.close()
    
        df2file.df2fileShell(
                             complicatedNamePart=f'{complicatedNamePart}_Temporal',
                             dfIn=df,
                             fileFormatChoice=fileFormatChoice,
                             method=method.split('.')[0] + method.split('.')[1].capitalize() if '.' in method else method, # чтобы избавиться от лишней точки в имени файла
                             coLabFolder=coLabFolder,
                             currentMoment=momentCurrent.strftime("%Y%m%d") # .strftime -- чтобы варьировать для итоговой директории и директории Temporal
                             )
        warnings.filterwarnings("ignore")
        print(
'Сейчас появится надпись: "An exception has occurred, use %tb to see the full traceback.\nSystemExit" -- так и должно быть',
'Модуль создан при финансовой поддержке Российского научного фонда по гранту 22-28-20473'
              )
        sys.exit()

    return df


# 1.2 для обработки выдачи аргумента fields
def fieldsProcessor(dfIn, fieldsColumn, response):
    df = dfIn.copy()
    idColumnS = []
    for column in df.columns:
    # for column in df.columns[1:]: # для отладки
        if 'id' in column:
            # print('column:', column) # для отладки
            idColumnS.append(column)
    columnsToJSON = varPreprocessor.jsonChecker(df)
    idColumnS.extend(columnsToJSON)
    
    fieldsDf = pandas.json_normalize(response[fieldsColumn])
    idS = pandas.json_normalize(response[fieldsColumn])['id'].to_list()
    if len(idS) > 0:
        idsCopy = pandas.json_normalize(response[fieldsColumn])['id'].to_list()
        idsCopyStr = '' # список в текстовый объект, чтобы ниже подать его внутрь столбца idColumnsConcatinated, созданного конкатенацией idColumnS
        for idCopy in idsCopy: idsCopyStr += str(idCopy) + ', '
        idsCopyStr = idsCopyStr[:-2]
        # print('idsCopyStr':, idsCopyStr) # для отладки
    
    def fieldsIdsChecker(cellContent): # функция, приминяемая ниже посредством apply , чтобы ускорить процесс (по сравнению с циклом по ячейкам)
        idsCopy = cellContent.split('idsCopy')[1].split(', ')
        cellContent = cellContent.split('idsCopy')[0]
        idsToItemS = []
        for idCopy in idsCopy:
            if idCopy in cellContent:
                idsToItemS.append(int(idCopy))
        if idsToItemS != []:
            # print('idsToItemS не пустой список:', idsToItemS) # для отладки
            try: return fieldsDf[fieldsDf['id'].isin(idsToItemS)].to_dict('records')
            except:
                print('!!! Ошибка:', sys.exc_info()[1])
                # print('dict:', fieldsDf[fieldsDf['id'].isin(idsToItemS)].to_dict('records')) # для отладки
                return ''
        else:
            # print('idsToItemS пустой список:', idsToItemS) # для отладки
            return ''

    df['idColumnsConcatinated'] = ''
    for idColumn in idColumnS:
        df['idColumnsConcatinated'] += ' ' + df[idColumn].astype(str)

    df['idColumnsConcatinated'] += 'idsCopy' + idsCopyStr
    df[fieldsColumn] = df['idColumnsConcatinated'].apply(fieldsIdsChecker)
    df[fieldsColumn] = df[fieldsColumn].replace('N/A', numpy.NaN)
    return df

# # Код, чтобы распарсить любой из двух столбцов датафрейма itemS с выдачей аргумента fields
# fieldsColumn = 'groups'
# # fieldsColumn = 'profiles'
# fieldsJSON = []
# for cell in itemS[fieldsColumn].dropna():
#     fieldsJSON.extend(cell)
# pandas.json_normalize(fieldsJSON).drop_duplicates('id').reset_index(drop=True)


# In[ ]:


# # 2 Авторская функция исполнения скрипта


# In[ ]:


def newsFeedSearch(
                   access_token=None,
                   count=None,
                   end_time=None,
                   fields=None,
                   latitude=None,
                   longitude=None,
                   params=None,
                   q=None,
                   returnDfs=False,
                   start_time=None
                   ):
    """
    Функция для выгрузки характеристик контента ВК методом его API newsfeed.search. Причём количество объектов выгрузки максимизируется путём её сегментирования по годам и месяцам

    Parameters
    ----------
    Аргументы этой функции аналогичны аргументам метода https://dev.vk.com/ru/method/newsfeed.search , за исключением аргументов params и returnDfs
    Причём они могут быть поданы и в качестве самостоятельных аргументов функции, и в качестве словаря params , который обычно подаётся в метод get пакета requests
    access_token : str
           count : int
        end_time : int
          fields : list
        latitude : int
       longitude : int
          params : dict -- в случае наличия готового словаря с аргументами метода https://dev.vk.com/ru/method/newsfeed.search , чтобы не подавать эти аргументы по отдельности
               q : str
       returnDfs : bool -- в случае True функция возвращает итоговый датафрейм с постами и их метаданными
      start_time : int
    """
    if (params == None) & (access_token == None) & (count == None) & (end_time == None) & (fields == None) & (latitude == None) & (longitude == None) & (q == None) & (start_time == None) & (returnDfs == False):
        # print('Пользователь не подал аргументы') # для отладки
        expiriencedMode = False
        count = 200
    else:
        expiriencedMode = True
        if params != None:
            access_token = params['access_token'] if 'access_token' in params.keys() else None
            q = params['q'] if 'q' in params.keys() else None

            if 'start_time' in params.keys():
                start_time = params['start_time']
                if type(start_time) == str: start_time = int(start_time)
            else: start_time = None

            if 'count' in params.keys():
                count = params['count']
                if type(count) == str: count = int(count)
            else: count = None

            if 'end_time' in params.keys():
                end_time = params['end_time']
                if type(end_time) == str: end_time = int(end_time)
            else: end_time = None

            fields = params['fields'] if 'fields' in params.keys() else None

            if 'latitude' in params.keys():
                latitude = params['latitude']
                if type(latitude) == str: latitude = int(latitude)
            else: latitude = None

            if 'longitude' in params.keys():
                longitude = params['longitude']
                if type(longitude) == str: longitude = int(longitude)
            else: longitude = None

    if expiriencedMode == False:
        print(
'''    Для исполнения скрипта не обязательны пререквизиты (предшествующие скрипты и файлы с данными). Но от пользователя требуется предварительно получить API key для авторизации в API ВК (см. примерную инструкцию: https://docs.google.com/document/d/1IiIWweiLP1GDl_f4yyhJO2F4K_RceTc3OSqMYotCXVg ). Для получения API key следует создать приложение и из него скопировать сервисный ключ. Приложение -- это как бы аккаунт для предоставления ему разных уровней авторизации (учётных данных, или Credentials) для доступа к содержимому ВК. Авторизация сервисным ключом позволяет использовать некоторые методы API -- в документации API ВК ( https://dev.vk.com/ru/method ) они помечены серым кружком (одним или в сочетании с кружками другого цвета). Его достаточно, если выполнять действия, которые были бы доступны Вам как обычному пользователю ВК: посмотреть открытые персональные и групповые страницы, почитать комментарии и т.п. Если же Вы хотите выполнить действия вроде удаления поста из чужого аккаунта, то Вам потребуется дополнительная авторизация.
    ВК может ограничить действие Вашего ключа или вовсе заблокировать его, если сочтёт, что Вы злоупотребляете автоматизированным доступом.'''
              )
    print(
'''    Скрипт нацелен на выгрузку характеристик контента ВК методом его API newsfeed.search. Причём количество объектов выгрузки максимизируется путём её сегментирования по годам и месяцам. При этом следует учесть, что текущая версия API ВК, на которой и основан скрипт, "не умеет" обращаться к контенту, опубликованному ранее 2020 года
    Для корректного исполнения скрипта просто следуйте инструкциям в возникающих по ходу его исполнения сообщениях. Скрипт исполняется и под MC OS, и под Windows.
    Преимущества скрипта перед выгрузкой контента из ВК вручную: гораздо быстрее, гораздо большее количество контента, его организация в формате таблицы Excel. Преимущества скрипта перед выгрузкой контента через непосредственно API ВК: гораздо быстрее, гораздо большее количество контента, не требуется тщательно изучать обширную и при этом неполную документацию методов API ВК'''
          )
    if expiriencedMode == False: input('--- После прочтения этой инструкции нажмите Enter')

# 2.0 Настройки и авторизация
# 2.0.0 Некоторые базовые настройки запроса к API ВК
    coLabFolder = coLabAdaptor.coLabAdaptor()
    fileFormatChoice = '.xlsx' # базовый формат сохраняемых файлов; формат .json добавляется опционально через наличие columnsToJSON
    folder = None
    folderFile = None
    goS = True
    itemS = pandas.DataFrame()
    keyOrder = 0
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС
    stageTarget = 0 # stageTarget принимает значения [0; 3] и относится к стадиям скрипта
    temporalName = None
    yearsRange = None

    momentCurrent = datetime.now() # запрос текущего момента
    print('\nТекущий момент:', momentCurrent.strftime("%Y%m%d_%H%M"), '-- он будет использована для формирования имён создаваемых директорий и файлов (во избежание путаницы в директориях и файлах при повторных запусках)\n')
    year = int(momentCurrent.strftime("%Y")) # в случае отсутствия пользовательского временнОго диапазона
        # с этого года возможно сегментирование по годам вглубь веков (пока выдача не пустая)
    yearMinByUser = None # в случае отсутствия пользовательского временнОго диапазона
    yearMaxByUser = None # в случае отсутствия пользовательского временнОго диапазона

# 2.0.1 Поиск следов прошлых запусков: ключей и данных; в случае их отсутствия -- получение настроек и (опционально) данных от пользователя
    rootNameS = os.listdir()
    # Поиск ключей
    if access_token == None:
        print('Проверяю наличие файла credentialsVK.txt с ключ[ом ами], гипотетически сохранённым[и] при первом запуске скрипта')
        if 'credentialsVK.txt' in rootNameS:
            file = open('credentialsVK.txt')
            API_keyS = file.read()
            print('Нашёл файл credentialsVK.txt; далее буду использовать ключ[и] из него:', API_keyS)
        else:
            print(
'''--- НЕ нашёл файл credentialsVK.txt . Введите в окно Ваш API key для авторизации в API ВК 
(примерная инструкция, как создать API key, доступна по ссылке https://docs.google.com/document/d/1dRqPGzLgr1wLp-_N6iuuZCmzCqrjYg1PuH7G7yomYdw ). Для подстраховки от ограничения действия API key желательно создать несколько ключей (три -- отлично) и ввести их без кавычек через запятую с пробелом
--- После ввода нажмите Enter'''
                  )
            while True:
                API_keyS = input()
                if len(API_keyS) != 0:
                    print('-- далее буд[е у]т использован[ы] эт[от и] ключ[и]')

                    from randan.tools.textPreprocessor import multispaceCleaner # авторский модуль для предобработки нестандартизированного текста
                    API_keyS = multispaceCleaner(API_keyS)
                    while API_keyS[-1] == ',': API_keyS = API_keyS[:-1] # избавиться от запятых в конце текста

                    file = open("credentialsVK.txt", "w+") # открыть на запись
                    file.write(API_keyS)
                    file.close()
                    break
                else:
                    print('--- Вы ничего НЕ ввели. Попробуйте ещё раз..')
        API_keyS = API_keyS.replace(' ', '') # контроль пробелов
        API_keyS = API_keyS.replace(',', ', ') # контроль пробелов
        API_keyS = API_keyS.split(', ')
    else: API_keyS = [access_token]
    print('Количество ключей:', len(API_keyS), '\n')

# 2.0.2 Скрипт может начаться с данных, сохранённых при прошлом исполнении скрипта, натолкнувшемся на ошибку
    # Поиск данных
    print('Проверяю наличие директории Temporal с данными и их мета-данными, гипотетически сохранёнными при прошлом запуске скрипта, натолкнувшемся на ошибку')
    for rootName in rootNameS:
        if 'Temporal' in rootName:
            if len(os.listdir(rootName)) == 7:
                file = open(f'{rootName}{slash}targetCount.txt')
                targetCount = file.read()
                file.close()
                targetCount = int(targetCount)
    
                file = open(f'{rootName}{slash}method.txt')
                method = file.read()
                file.close()
    
                file = open(f'{rootName}{slash}year.txt')
                year = file.read()
                file.close()
                year = int(year)
    
                file = open(f'{rootName}{slash}q.txt', encoding='utf-8') # 
                q = file.read()
                file.close()
                if q == '': q = None # для единообразия
    
                file = open(f'{rootName}{slash}yearsRange.txt')
                yearsRange = file.read()
                file.close()
    
                file = open(f'{rootName}{slash}stageTarget.txt')
                stageTarget = file.read()
                file.close()
                stageTarget = int(stageTarget)
    
                print(f'Нашёл директорию "{rootName}". В этой директории следующие промежуточные результаты одного из прошлых запусков скрипта:'
                      # , '\n- было выявлено целевое число объектов (targetCount)', targetCount
                      , '\n- скрипт остановился на методе', method)
                if year < int(momentCurrent.strftime("%Y")): print('- и на годе (при сегментировани по годам)', year)
                print('- пользователь НЕ сформулировал запрос-фильтр' if q == None else  f'- пользователь сформулировал запрос-фильтр как "{q}"')
                print('- пользователь НЕ ограничил временнОй диапазон' if yearsRange == None else  f'- пользователь ограничил временнОй диапазон границами {yearsRange}')
                print(
'''--- Если хотите продолжить дополнять эти промежуточные результаты, нажмите Enter
--- Если эти промежуточные результаты уже не актуальны и хотите их удалить, введите "R" и нажмите Enter
--- Если хотите найти другие промежуточные результаты, нажмите пробел и затем Enter'''
                  )
                decision = input()
                if len(decision) == 0:
                    temporalNameS = os.listdir(rootName)
                    for temporalName in temporalNameS:
                        if '.xlsx' in temporalName: break
                    itemS = pandas.read_excel(f'{rootName}{slash}{temporalName}', index_col=0)
    
                    for temporalName in temporalNameS:
                        if '.json' in temporalName:
                            itemS = itemS.merge(pandas.read_json(f'{rootName}{slash}{temporalName}'), on='id', how='outer')
                            break
    
                    if yearsRange != None:
                        yearsRange = yearsRange.split('-')
                        yearMaxByUser, yearMinByUser, yearsRange = calendarWithinYear.yearsRangeParser(yearsRange)
# Данные, сохранённые при прошлом запуске скрипта, загружены; их метаданные (q, yearsRange, stageTarget) будут использоваться при исполнении скрипта
                    break
                elif decision == 'R': shutil.rmtree(rootName, ignore_errors=True)

# 2.0.3 Если такие данные, сохранённые при прошлом запуске скрипта, не найдены, возможно, пользователь хочет подать свои данные для их дополнения
    if temporalName == None: # если itemsTemporal, в т.ч. пустой, не существует
            # и, следовательно, не существуют данные, сохранённые при прошлом запуске скрипта, натолкнувшемся на ошибку
        rootName = 'No folder'
        print('Не найдены подходящие данные, гипотетически сохранённые при прошлом запуске скрипта, натолкнувшемся на ошибку')
        print(
'''
Возможно, Вы располагаете файлом, в котором есть ранее выгруженные из ВК методом newsfeed.search данные, и который хотели бы дополнить? Или планируете первичный сбор контента?
--- Если планируете первичный сбор, нажмите Enter
--- Если располагаете файлом формата XLSX, укажите полный путь, включая название файла, и нажмите Enter. Затем при необходимости сможете добавить к нему другие располагаемые файлы'''
              )
        while True:
            folderFile = input()
            if len(folderFile) == 0:
                folderFile = None # для унификации
                break
            else:
                itemS, error, folder = files2df.files2df(folderFile)
                if error != None:
                    if 'No such file or directory' in error:
                        print('Путь:', folderFile, '-- не существует; попробуйте, пожалуйста, ещё раз..')
                else: break
            # display(itemS)
# Теперь определены объекты: folder и folderFile (оба None или пользовательские), itemS (пустой или с прошлого запуска, или пользовательский), slash

# 2.0.4 Пользовательские настройки запроса к API ВК
        if q == None: # если пользователь не подал этот аргумент в рамках experiencedMode
            print(
'''Скрипт умеет искать контент в постах открытых аккаунтов по текстовому запросу-фильтру
--- Введите текст запроса-фильтра, который ожидаете найти в постах, после чего нажмите Enter'''
                  )
            if folderFile != None:
                print(
'ВАЖНО! В результате исполнения текущего скрипта данные из указанного Вами файла', folderFile, 'будут дополнены актуальными данными из выдачи скрипта',
'(возможно появление новых объектов и новых столбцов, а также актуализация содержимого столбцов),',
'поэтому, вероятно, следует ввести тот же запрос-фильтр, что и при формировании указанного Вами файла'
                      )
            q = input()
            if q == '': q = None # для единообразия
            else: print('')

        # Ограничения временнОго диапазона
        if (start_time == None) & (end_time == None) & (yearsRange == None): # если пользователь не подал эти аргументы в рамках experiencedMode
            print(
'''Если требуется конкретный временнОй диапазон, то можно использовать его не на текущем этапе выгрузки данных, а на следующем этапе -- предобработки датафрейма с выгруженными данными. Проблема в том, что без назначения временнОго диапазона метод newsfeed.search выдаёт ограниченное количество объектов, причём наиболее приближенных к текущему моменту
--- Поэтому если всё же требуется назначить временнОй диапазон на этапе выгрузки данных, назначьте его годами (а не более мелкими единицами времени). Для назначения диапазона введите без кавычек минимальный год диапазона, тире, максимальный год диапазона (минимум и максимум могут совпадать в такой записи: "год-тот же год") и нажмите Enter !!! Cледует учесть, что текущая версия API ВК "не умеет" обращаться к контенту, опубликованному ранее 2020 года
--- Если НЕ требуется назначить временнОй диапазон на этапе выгрузки данных, нажмите Enter'''
                  )
            while True:
                yearsRange = input()
                if len(yearsRange) != 0:
                    yearsRange = re.sub(r' *', '', yearsRange)
                    if '-' in yearsRange:
                        yearsRange = yearsRange.split('-')
                        if len(yearsRange) == 2:
                            if (len(yearsRange[0]) == 4) & (len(yearsRange[1]) == 4):
                                yearMaxByUser, yearMinByUser, yearsRange = calendarWithinYear.yearsRangeParser(yearsRange)
                                year = yearMaxByUser
                                break
                            else: print('--- Вы ввели год[ы] НЕ из четырёх цифр. Попробуйте ещё раз..')
                        else: print('--- Вы ввели тире, но при этом ввели НЕ два года. Попробуйте ещё раз..')
                    else: print('--- Вы НЕ ввели тире. Попробуйте ещё раз..')
                else:
                    yearsRange = None # для унификации
                    break
        if start_time != None:
            yearMinByUser = int(datetime.fromtimestamp(start_time).strftime('%Y')) # из experiencedMode
            # print('elif start_time != None:', yearMinByUser) # для отладки

        if end_time != None:
            yearMaxByUser = int(datetime.fromtimestamp(end_time).strftime('%Y')) # из experiencedMode
            # print('elif end_time != None:', yearMaxByUser) # для отладки
            year = yearMaxByUser

        if (yearMinByUser != None) & (yearMaxByUser == None): yearMaxByUser = int(momentCurrent.strftime("%Y")) # в случае отсутствия пользовательской верхней временнОй границы при наличии нижней
        elif (yearMinByUser == None) & (yearMaxByUser != None): yearMaxByUser = 1970 # в случае отсутствия пользовательской нижней временнОй границы при наличии верхней

        # print('yearMinByUser', yearMinByUser) # для отладки
        # print('yearMaxByUser', yearMaxByUser) # для отладки

        if (start_time == None) & (yearMinByUser != None): start_time = int(datetime(yearMinByUser, 1, 1).timestamp()) # int(time.mktime(datetime(yearMinByUser, 1, 1).timetuple()))
        if (end_time == None) & (yearMaxByUser != None): end_time = int(datetime(yearMaxByUser, 12, 31).timestamp())

        if yearsRange != None: print('') # чтобы был отступ, если пользователь подал этот аргумент

        if count == None: count = 200 # если ни пользователь не подал, ни в данных, сохранённых при прошлом запуске скрипта, нет

# Сложная часть имени будущих директорий и файлов
    complicatedNamePart = '_VK'
    if q != None: complicatedNamePart += "_" + q if len(q) < 50 else "_" + q[:50]
    complicatedNamePart += "" if ((yearMinByUser == None) & (yearMaxByUser == None)) else "_" + str(yearMinByUser) + '-' + str(yearMaxByUser)
    # print('complicatedNamePart', complicatedNamePart)

# 2.1 Первичный сбор контента методом search
# 2.1.0 Первое обращение к API БЕЗ аргументов start_time, end_time (этап stage = 0)
    stage = 0
    method = 'newsfeed.search'
    iteration = 0 # номер итерации применения текущего метода
    pause = 0.25
    print(
f'В скрипте используются следующие аргументы метода {method} API ВК: q, start_from, start_time, end_time, expand.',
'Эти аргументы пользователю скрипта лучше не кастомизировать во избежание поломки скрипта.',
f'Если хотите добавить другие аргументы метода {method} API ВК, доступные по ссылке https://dev.vk.com/ru/method/newsfeed.search ,',
f'-- можете подать их в скобки функции newsFeedSearch перед её запуском или скопировать код исполняемого сейчас скрипта и сделать это внутри кода внутри метода {method} в разделе 2'
          )
    # print('expiriencedMode:', expiriencedMode) # для отладки
    if expiriencedMode == False: input('--- После прочтения этой инструкции нажмите Enter')
    print('') # для отступа

    if stage >= stageTarget: # eсли нет временного файла stage.txt с указанием пропустить этап
        print('Первое обращение к API -- прежде всего, чтобы узнать примерное число доступных релевантных объектов')
        # print('    start_from', start_from) # для отладки
        itemsAdditional, goS, iteration, keyOrder, pause, response = bigSearch(
                                                                               API_keyS=API_keyS,
                                                                               count=count,
                                                                               end_time=end_time,
                                                                               fields=fields,
                                                                               iteration=iteration,
                                                                               keyOrder=keyOrder,
                                                                               latitude=latitude,
                                                                               longitude=longitude,
                                                                               pause=pause,
                                                                               q=q,
                                                                               start_from=None,
                                                                               start_time=start_time
                                                                               )
        targetCount = response['total_count']
        if targetCount == 0:
            print(
'  Искомых объектов на серверах ВК по Вашему запросу, увы, ноль, поэтому нет смысла в продолжении исполнения скрипта. Что делать? Поменяйте настройки запроса и запустите скрипт с начала'
                  )
            warnings.filterwarnings("ignore")
            print(
'Сейчас появится надпись: "An exception has occurred, use %tb to see the full traceback.\nSystemExit" -- так и должно быть',
'Модуль создан при финансовой поддержке Российского научного фонда по гранту 22-28-20473'
                  )
            sys.exit()

        # if len(itemS) < targetCount: # на случай достаточности
        itemS = dfsProcessor(
                             complicatedNamePart=complicatedNamePart,
                             coLabFolder=coLabFolder,
                             fileFormatChoice=fileFormatChoice,
                             goS=goS,
                             dfAdd=itemsAdditional,
                             dfFinal=itemS,
                             dfIn=itemS,
                             method=method,
                             momentCurrent=momentCurrent,
                             q=q,
                             slash=slash,
                             stage=stage,
                             targetCount=targetCount,
                             year=year,
                             yearsRange=yearsRange
                             )
        print('  Проход по всем следующим страницам с выдачей          ')
        while 'next_from' in response.keys():
            start_from = response['next_from']
            # print('    start_from', start_from) # для отладки
            itemsAdditional, goS, iteration, keyOrder, pause, response = bigSearch(
                                                                                   API_keyS=API_keyS,
                                                                                   count=count,
                                                                                   end_time=end_time,
                                                                                   fields=fields,
                                                                                   iteration=iteration,
                                                                                   keyOrder=keyOrder,
                                                                                   latitude=latitude,
                                                                                   longitude=longitude,
                                                                                   pause=pause,
                                                                                   q=q,
                                                                                   start_from=start_from,
                                                                                   start_time=start_time
                                                                                   )
            # print('''    response['next_from'] после bigSearch''', response['next_from']) # для отладки
            itemS = dfsProcessor(
                                 complicatedNamePart=complicatedNamePart,
                                 coLabFolder=coLabFolder,
                                 fileFormatChoice=fileFormatChoice,
                                 goS=goS,
                                 dfAdd=itemsAdditional,
                                 dfFinal=itemS,
                                 dfIn=itemS,
                                 method=method,
                                 momentCurrent=momentCurrent,
                                 q=q,
                                 slash=slash,
                                 stage=stage,
                                 targetCount=targetCount,
                                 year=year,
                                 yearsRange=yearsRange
                                 )
        print('  Искомых объектов', targetCount, ', а найденных БЕЗ сегментирования по годам и месяцам:', len(itemS))

# 2.1.1 Этап сегментирования по годам и месяцам (stage = 1)
    stage = 1
    if stage >= stageTarget: # eсли нет временного файла stage.txt с указанием пропустить этап
        if len(itemS) < targetCount:
        # -- для остановки алгоритма, если все искомые объекты найдены БЕЗ сегментирования по годам и месяцам
            print(
'''Увы, без назначения временнОго диапазона метод newsfeed.search выдаёт ограниченное количество объектов, причём наиболее приближенных к текущему моменту. Поэтому внутри каждого года, начиная с текущего, помесячно выгружаю контент, после чего меняю год -- вглубь веков, пока не достигну заданной пользователем левой границы временнОго диапазона или года с пустой выдачей'''
                  )
            print(
'''--- Если хотите для поиска дополнительных объектов попробовать сегментирование по годам и месяцам, просто нажмите Enter, но учтите, что поиск может занять минуты и даже часы
--- Если НЕ хотите, нажмите пробел и затем Enter'''
                  )
            if len(input()) == 0:
                while True:
                    # print('Ищу текст запроса-фильтра в контенте за', year, 'год')
                    calendar = calendarWithinYear.calendarWithinYear(year)
                    itemsYearlyAdditional = pandas.DataFrame()
                    calendarColumnS = calendar.columns
                    if year == int(momentCurrent.strftime("%Y")): calendarColumnS = calendarColumnS[:int(momentCurrent.strftime("%m"))]
                            # чтобы исключить проход по будущим месяцам текущего года
                    for month in calendarColumnS:
                        print('Ищу текст запроса-фильтра в контенте за',  month, 'месяц', year, 'года', '               ') # , end='\r'
                        print('  Заход на первую страницу выдачи', '               ', end='\r')
                        itemsMonthlyAdditional, goS, iteration, keyOrder, pause, response = bigSearch(
                                                                                                      API_keyS=API_keyS,
                                                                                                      count=count,
                                                                                                      end_time=int(datetime(year, int(month), int(calendar[month].dropna().index[-1])).timestamp()),
                                                                                                      fields=fields,
                                                                                                      iteration=iteration,
                                                                                                      keyOrder=keyOrder,
                                                                                                      latitude=latitude,
                                                                                                      longitude=longitude,
                                                                                                      pause=pause,
                                                                                                      q=q,
                                                                                                      start_from=None,
                                                                                                      start_time=int(datetime(year, int(month), 1).timestamp())
                                                                                                      )
                        # print('itemsMonthlyAdditional:') # для отладки
                        # display(itemsMonthlyAdditional.sort_values('date')['date'].drop_duplicates()) # для отладки
                        itemsYearlyAdditional = dfsProcessor(
                                                             complicatedNamePart=complicatedNamePart,
                                                             coLabFolder=coLabFolder,
                                                             fileFormatChoice=fileFormatChoice,
                                                             goS=goS,
                                                             dfAdd=itemsMonthlyAdditional,
                                                             dfFinal=itemS,
                                                             dfIn=itemsYearlyAdditional,
                                                             method=method,
                                                             momentCurrent=momentCurrent,
                                                             q=q,
                                                             slash=slash,
                                                             stage=stage,
                                                             targetCount=targetCount,
                                                             year=year,
                                                             yearsRange=yearsRange
                                                             )
                        # print('len(itemsYearlyAdditional):', len(itemsYearlyAdditional)) # для отладки

                        print('  Проход по всем следующим страницам с выдачей', '               ', end='\r')
                        while 'next_from' in response.keys():
                            start_from = response['next_from']
                            # print('    start_from', start_from) # для отладки
                            itemsMonthlyAdditional, goS, iteration, keyOrder, pause, response = bigSearch(
                                                                                                          API_keyS=API_keyS,
                                                                                                          count=count,
                                                                                                          end_time=int(datetime(year, int(month), int(calendar[month].dropna().index[-1])).timestamp()),
                                                                                                          fields=fields,
                                                                                                          iteration=iteration,
                                                                                                          keyOrder=keyOrder,
                                                                                                          latitude=latitude,
                                                                                                          longitude=longitude,
                                                                                                          pause=pause,
                                                                                                          q=q,
                                                                                                          start_from=start_from,
                                                                                                          start_time=int(datetime(year, int(month), 1).timestamp())
                                                                                                          )
                            # print('itemsMonthlyAdditional:') # для отладки
                            # display(itemsMonthlyAdditional.sort_values('date')['date'].drop_duplicates()) # для отладки
                            itemsYearlyAdditional = dfsProcessor(
                                                                 complicatedNamePart=complicatedNamePart,
                                                                 coLabFolder=coLabFolder,
                                                                 fileFormatChoice=fileFormatChoice,
                                                                 goS=goS,
                                                                 dfAdd=itemsMonthlyAdditional,
                                                                 dfFinal=itemS,
                                                                 dfIn=itemsYearlyAdditional,
                                                                 method=method,
                                                                 momentCurrent=momentCurrent,
                                                                 q=q,
                                                                 slash=slash,
                                                                 stage=stage,
                                                                 targetCount=targetCount,
                                                                 year=year,
                                                                 yearsRange=yearsRange
                                                                 )
                            # print('len(itemsYearlyAdditional):', len(itemsYearlyAdditional)) # для отладки
                            time.sleep(pause)
                    # print('itemsYearlyAdditional:') # для отладки
                    # display(itemsYearlyAdditional.sort_values('date')['date'].drop_duplicates()) # для отладки
                    itemS = dfsProcessor(
                                         complicatedNamePart=complicatedNamePart,
                                         coLabFolder=coLabFolder,
                                         fileFormatChoice=fileFormatChoice,
                                         goS=goS,
                                         dfAdd=itemsYearlyAdditional,
                                         dfFinal=itemS,
                                         dfIn=itemS,
                                         method=method,
                                         momentCurrent=momentCurrent,
                                         q=q,
                                         slash=slash,
                                         stage=stage,
                                         targetCount=targetCount,
                                         year=year,
                                         yearsRange=yearsRange
                                         )
                    # print('len(itemS):', len(itemS)) # для отладки
                    # display(itemS.head())
                    # print('Число столбцов:', itemS.shape[1], ', число строк', itemS.shape[0])

                    if len(itemsYearlyAdditional) == 0:
                        print(f'\nВыдача для года {year} -- пуста'
                              , '\n--- Если НЕ хотите для поиска дополнительных объектов попробовать двигаться к следующему месяцу вглубь веков, просто нажмите Enter'
                              , '\n--- Если хотите, нажмите пробел и затем Enter')
                        if len(input()) == 0:
                            # print(f'\nЗавершил проход по заданному пользователем временнОму диапазону: {yearMinByUser}-{yearMaxByUser}')
                            break

                    elif yearMinByUser != None: # если пользователь ограничил временнОй диапазон
                        if year <= yearMinByUser:
                            print(f'Завершил проход по заданному пользователем временнОму диапазону: {yearMinByUser}-{yearMaxByUser}')
                            break

                    print('  Искомых объектов', targetCount, ', а найденных после добавления контента', year, 'года:', len(itemS), '                    ')
                    year -= 1
                print('Искомых объектов', targetCount, ', а найденных:', len(itemS), '          ')

        # pandas.set_option('display.max_columns', None)
        display(itemS.head())
        print('Число столбцов:', itemS.shape[1], ', число строк', itemS.shape[0])

    elif stage < stageTarget:
        print(f'\nЭтап {stage} пропускаю согласно настройкам из файла stage.txt в директории "{momentCurrent.strftime("%Y%m%d")}{complicatedNamePart}_Temporal"')

# 2.1.2 Экспорт выгрузки метода search и финальное завершение скрипта
    df2file.df2fileShell(
                         complicatedNamePart=complicatedNamePart,
                         dfIn=itemS,
                         fileFormatChoice=fileFormatChoice,
                         method=method.split('.')[0] + method.split('.')[1].capitalize() if '.' in method else method, # чтобы избавиться от лишней точки в имени файла
                         coLabFolder=coLabFolder,
                         currentMoment=momentCurrent.strftime("%Y%m%d_%H%M") # .strftime -- чтобы варьировать для итоговой директории и директории Temporal
                         )

    print('Скрипт исполнен. Модуль создан при финансовой поддержке Российского научного фонда по гранту 22-28-20473')
    if os.path.exists(rootName):
        print('rootName:', rootName)
        print(
'Поскольку данные, сохранённые при одном из прошлых запусков скрипта в директорию Temporal, успешно использованы, УДАЛЯЮ её во избежание путаницы при следующих запусках скрипта'
              )
        shutil.rmtree(rootName, ignore_errors=True)
    if fields != None: print(
'''
Чтобы распаковать JSON из любого столбца, содержащего этот формат, в отдельный датафрейм, используйте такой код:
import pandas
column = 'Имя_столбца'
JSONS = []
for cellContent in Исходный_датафрейм[column].dropna():
    JSONS.extend(cellContent)
Новый_датафрейм = pandas.json_normalize(JSONS).drop_duplicates('id').reset_index(drop=True)
'''
                             )
    if returnDfs: return itemS
# warnings.filterwarnings("ignore")
# print('Сейчас появится надпись: "An exception has occurred, use %tb to see the full traceback.\nSystemExit" -- так и должно быть')
# input()
# sys.exit()



import pandas as pd
import numpy as np

from pandas.api.types import is_numeric_dtype
from .utils import get_categories

from statsmodels.stats.diagnostic import lilliefors
from scipy.stats import shapiro

from IPython.display import display

class NominalStatistics:
    """
    A class producing descriptive statistics relevant for nominal variables.

    Parameters
    ----------
    data : pd.DataFrame
        Data used to perform the analysis
    variables : list
        Variables from data to include in the analysis
    frequencies : bool
        Whether to show frequency tables
    show_results : bool 
        Whether to show results of analysis
    n_decimals : int 
        Number of digits to round results when showing them

    """
    def __init__(
        self, 
        data, 
        variables=None,
        frequencies=True,
        show_results=True,
        n_decimals=3
    ):
        
        if variables is not None:
        
            if not isinstance(variables, list):
                phrase = 'Variables should be passed as list. Type {} was passed instead.'
                raise TypeError(phrase.format(type(variables)))
            
            else:
                self._data = data[variables].copy()
        
        else:
            self._data = data.copy()
        
        self._variables = list(self._data.columns)
        
        if show_results:
            self.show_results(n_decimals=n_decimals)
        if frequencies:
            self.show_frequencies()
    
    def show_results(self, n_decimals=3):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them
        """
        print('\nNOMINAL STATISTICS SUMMARY')
        print('------------------\n')
        display(self.summary().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("method .summary()"))
        if len(self._mult_modes) > 0:
            vars_ = ', '.join(self._mult_modes)
            print(f'Following variables have multiple modes: {vars_}')
    
    def show_frequencies(self, n_decimals=3):
        """
        Show frequency tables.
        
        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them
        """
        print('\nFREQUENCIES')
        for var in self._variables:
            print('------------------\n')
            print(f'variable: {var}')
            display(self.frequencies()[var].style\
                        .format(None, na_rep="", precision=n_decimals)\
                        .set_caption(f"method .frequencies()['{var}']"))
    
    def _get_statistics(self):
        measures = {}
        self._mult_modes = []
        for var in self._variables:
            ser = self._data[var]
            n = len(ser.dropna())
            mode = ser.mode()
            if len(mode) > 1:
                self._mult_modes.append(var)
            mode = mode.iloc[0]
            entr = NominalStatistics._entropy_coef(ser)
            cqv = NominalStatistics._cqv_coef(ser)
            measures.update({var: [n, mode, entr, cqv]})
        measures = pd.DataFrame(measures, index=['N', 'mode', 'entropy coef.', 'quality var.'])
        return measures.T
    
    @staticmethod
    def _entropy_coef(series):
        p = series.value_counts(normalize=True) 
        entr_obs = (p * p.apply(np.log)).sum()
        n = len(p)
        p_exp = np.array([1/n] * n)
        entr_exp = (p_exp * np.log(p_exp)).sum()
        coef = entr_obs / entr_exp
        return coef
    
    @staticmethod
    def _cqv_coef(series):
        n = len(series)
        p = series.value_counts()
        k = len(p)
        p_sq_sum = (p ** 2).sum()
        coef = (k * (n**2 - p_sq_sum)) / ((n**2) * (k - 1))
        return coef
    
    def summary(self):
        """
        Return aggregated results of the analysis.
        """
        return self._get_statistics()
    
    def __repr__(self):
        n_vars_ = len(self._variables)
        return f'<NominalStatistics Object for {n_vars_} variables>'
    
    @staticmethod
    def _get_frequencies(series):
        raw = series.value_counts()
        raw.name = 'N'
        normalized = series.value_counts(normalize=True) * 100
        normalized.name = '%'
        return pd.concat([raw, normalized], axis=1)
    
    def frequencies(self):
        """
        Return a dictionary of all frequency tables. 
        To get a particular frequency table, use a variable's name as a key of the dictionary. 
        """
        freq = {}
        for var in self._variables:
            ser = self._data[var]
            freq.update({var: NominalStatistics._get_frequencies(ser)})
        return freq

class OrdinalStatistics:
    """
    A class producing descriptive statistics relevant for ordinal variables.

    Parameters
    ----------
    data : pd.DataFrame
        Data used to perform the analysis
    variables : list
        Variables from data to include in the analysis
    frequencies : bool
        Whether to show frequency tables
    show_results : bool 
        Whether to show results of analysis
    n_decimals : int 
        Number of digits to round results when showing them

    """
    def __init__(
        self, 
        data, 
        variables=None,
        frequencies=True,
        show_results=True,
        n_decimals=3
    ):
        
        if variables is not None:
        
            if not isinstance(variables, list):
                phrase = 'Variables should be passed as list. Type {} was passed instead.'
                raise TypeError(phrase.format(type(variables)))
            
            else:
                self._data = data[variables].copy()
        
        else:
            self._data = data.copy()
        
        self._variables = list(self._data.columns)
        self._mappers = self._get_mappers_for_nonumerical_vars()
        
        if len(self._mappers) > 0:
            for var in self._mappers.keys():
                self._data[var] = self._data[var].map(self._mappers[var][0]).astype(float)
        
        if show_results:
            self.show_results(n_decimals=n_decimals)
        if frequencies:
            self.show_frequencies()
    
    def _get_mappers_for_nonumerical_vars(self):
        nonnum_vars = [var for var in self._variables if not is_numeric_dtype(self._data[var])]
        mappers = {}
        for var in nonnum_vars:
            mappers.update({var: OrdinalStatistics._get_mappers_for_one_var(self._data[var])})
        return mappers
    
    @staticmethod
    def _get_mappers_for_one_var(series):
        categories = get_categories(series)
        numbers = [float(num) for num in list(range(len(categories)))]
        dir_mapper = dict(zip(categories, numbers))
        inv_mapper = dict(zip(numbers, categories))
        return [dir_mapper, inv_mapper]
    
    def show_results(self, n_decimals=3):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them
        """
        print('\nORDINAL STATISTICS SUMMARY')
        print('------------------\n')
        display(self.summary().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("method .summary()"))
        if len(self._mult_modes) > 0:
            vars_ = ', '.join(self._mult_modes)
            print(f'Following variables have multiple modes: {vars_}')
    
    def show_frequencies(self, n_decimals=3):
        """
        Show frequency tables.
        
        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them
        """
        print('\nFREQUENCIES')
        for var in self._variables:
            print('------------------\n')
            print(f'variable: {var}')
            display(self.frequencies()[var].style\
                        .format(None, na_rep="", precision=n_decimals)\
                        .set_caption(f"method .frequencies()['{var}']"))
    
    def _get_statistics(self):
        measures = {}
        self._mult_modes = []
        for var in self._variables:
            ser = self._data[var]
            n = len(ser.dropna())
            mode = ser.mode()
            if len(mode) > 1:
                self._mult_modes.append(var)
            mode = mode.iloc[0]
            #display(ser)
            median = ser.median()
            q25 = ser.quantile(0.25)
            q75 = ser.quantile(0.75)
            min_ = ser.min()
            max_ = ser.max()
            range_ = max_ - min_
            entr = OrdinalStatistics._entropy_coef(ser)
            cqv = OrdinalStatistics._cqv_coef(ser)
            iqv = q75 - q25
            iqv_norm = iqv / range_
            measures.update({var: [n, mode, median, q25, q75, iqv, iqv_norm, min_, max_, range_, entr, cqv]})
        measures = pd.DataFrame(measures,
                                index=['N', 'mode', 'median',
                                       '25%', '75%', 'interquart. range',
                                       'interquart. range (norm.)',
                                       'min', 'max', 'range',
                                       'entropy coef.', 'quality var.'])
        if len(self._mappers) > 0:
            for var in self._mappers.keys():
                measures.loc[['mode', 'median', '25%', '75%', 'min', 'max'], var] = \
                measures.loc[['mode', 'median', '25%', '75%', 'min', 'max'], var].map(self._mappers[var][1])
        return measures.T
    
    @staticmethod
    def _entropy_coef(series):
        p = series.value_counts(normalize=True) 
        entr_obs = (p * p.apply(np.log)).sum()
        n = len(p)
        p_exp = np.array([1/n] * n)
        entr_exp = (p_exp * np.log(p_exp)).sum()
        coef = entr_obs / entr_exp
        return coef
    
    @staticmethod
    def _cqv_coef(series):
        n = len(series)
        p = series.value_counts()
        k = len(p)
        p_sq_sum = (p ** 2).sum()
        coef = (k * (n**2 - p_sq_sum)) / ((n**2) * (k - 1))
        return coef
    
    def summary(self):
        """
        Return aggregated results of the analysis.
        """
        return self._get_statistics()
    
    def __repr__(self):
        n_vars_ = len(self._variables)
        return f'<OrdinalStatistics Object for {n_vars_} variables>'
    
    @staticmethod
    def _get_frequencies(series):
        raw = series.value_counts()
        raw.name = 'N'
        normalized = series.value_counts(normalize=True) * 100
        normalized.name = '%'
        return pd.concat([raw, normalized], axis=1)
    
    def frequencies(self):
        """
        Return a dictionary of all frequency tables. 
        To get a particular frequency table, use a variable's name as a key of the dictionary. 
        """
        freq = {}
        for var in self._variables:
            ser = self._data[var]
            freqs = OrdinalStatistics._get_frequencies(ser)
            if var in self._mappers:
                freqs.index = freqs.index.map(self._mappers[var][1])
            freq.update({var: freqs})
        return freq

class ScaleStatistics:
    """
    A class producing descriptive statistics relevant for scale (a.k.a. interval) variables.

    Parameters
    ----------
    data : pd.DataFrame
        Data used to perform the analysis
    variables : list
        Variables from data to include in the analysis
    frequencies : bool
        Whether to show frequency tables
    normality_test : bool
        Whether to perform a normality test
    normality_test_type : str
        Which normality test to use. Available values: 'ks' (Kolmogorov-Smirnov's test) or 'sw' (Shapiro-Wilk' test)
    show_results : bool 
        Whether to show results of analysis
    n_decimals : int 
        Number of digits to round results when showing them

    """
    def __init__(
        self, 
        data, 
        variables=None,
        frequencies=False,
        normality_test=False,
        normality_test_type='ks',
        show_results=True,
        n_decimals=3
    ):
        
        if variables is not None:
        
            if not isinstance(variables, list):
                phrase = 'Variables should be passed as list. Type {} was passed instead.'
                raise TypeError(phrase.format(type(variables)))
            
            else:
                self._data = data[variables].copy()
        
        else:
            self._data = data.copy()
        
        self._variables = list(self._data.columns)
        self._mappers = self._get_mappers_for_nonumerical_vars()
        
        self.normality_test_type = normality_test_type
        
        if len(self._mappers) > 0:
            for var in self._mappers.keys():
                self._data[var] = self._data[var].map(self._mappers[var][0]).astype(float)
        
        if show_results:
            if len(self._mappers)>0:
                print('\nENCODING INFORMATION')
                print('------------------\n')
                print('Some of the variables are presented as categorical ones.')
                print('They were encoded according to the following rules:')
                for var in self._mappers:
                    print('------------------\n')
                    print(f'variable: {var}')
                    display(pd.DataFrame(self._mappers[var][0], index=['Encoded value']).T)
                    
            self.show_results(n_decimals=n_decimals)
        if normality_test:
            self.show_normality_test(n_decimals=n_decimals)
        if frequencies:
            self.show_frequencies()
            
        
    
    def normality_test(self, test_type='ks'):
        """
        Perform normality tests for all included variables.
        
        Parameters
        ----------
        test_type : str
        Which normality test to use. Available values: 'ks' (Kolmogorov-Smirnov's test) or 'sw' (Shapiro-Wilk' test)
        """
        if test_type not in ['ks', 'sw']:
            raise ValueError("Unknown normality test type. Possible values: 'ks' (Kolmogorov-Smirnov) ans 'sw' (Shapiro-Wilk)")
        
        results = {}
        
        for var in self._variables:
            ser = self._data[var]
            if test_type=='ks':
                stat, pval = lilliefors(ser.dropna(), pvalmethod='approx')
            elif test_type=='sw':
                stat, pval = shapiro(ser.dropna())
            results.update({var: [stat, pval]})
        
        results = pd.DataFrame(results, index=['statistic', 'p-value'])
        
        return results.T
    
    def show_normality_test(self, n_decimals=3):
        """
        Show results of normality tests for all included variables.
        
        Parameters
        ----------
        n_decimals : int 
        Number of digits to round results when showing them
        """
        print('\nNORMALITY TESTS')
        print('------------------\n')
        display(self.normality_test(self.normality_test_type).style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(f"method .normality_test(test_type='{self.normality_test_type}')"))
    
    def _get_mappers_for_nonumerical_vars(self):
        nonnum_vars = [var for var in self._variables if not is_numeric_dtype(self._data[var])]
        mappers = {}
        for var in nonnum_vars:
            mappers.update({var: ScaleStatistics._get_mappers_for_one_var(self._data[var])})
        return mappers
    
    @staticmethod
    def _get_mappers_for_one_var(series):
        categories = get_categories(series)
        numbers = [float(num) for num in list(range(len(categories)))]
        dir_mapper = dict(zip(categories, numbers))
        inv_mapper = dict(zip(numbers, categories))
        return [dir_mapper, inv_mapper]
    
    def show_results(self, n_decimals=3):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them
        """
        print('\nSCALE STATISTICS SUMMARY')
        print('------------------\n')
        display(self.summary().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption("method .summary()"))
        if len(self._mult_modes) > 0:
            vars_ = ', '.join(self._mult_modes)
            print(f'Following variables have multiple modes: {vars_}')
    
    def show_frequencies(self, n_decimals=3):
        """
        Show frequency tables.
        
        Parameters
        ----------
        n_decimals : int
            Number of digits to round results when showing them
        """
        print('\nFREQUENCIES')
        for var in self._variables:
            print('------------------\n')
            print(f'variable: {var}')
            display(self.frequencies()[var].style\
                        .format(None, na_rep="", precision=n_decimals)\
                        .set_caption(f"method .frequencies()['{var}']"))
    
    def _get_statistics(self):
        measures = {}
        self._mult_modes = []
        for var in self._variables:
            ser = self._data[var]
            n = len(ser.dropna())
            mode = ser.mode()
            if len(mode) > 1:
                self._mult_modes.append(var)
            mode = mode.iloc[0]
            #display(ser)
            median = ser.median()
            mean = ser.mean()
            q25 = ser.quantile(0.25)
            q75 = ser.quantile(0.75)
            min_ = ser.min()
            max_ = ser.max()
            range_ = max_ - min_
            std = ser.std()
            var_ = ser.var()
            entr = ScaleStatistics._entropy_coef(ser)
            cqv = ScaleStatistics._cqv_coef(ser)
            iqv = q75 - q25
            iqv_norm = iqv / range_
            measures.update({var: [n, mode, median, mean, q25, q75, iqv, iqv_norm, min_, max_, range_, std, var_, entr, cqv]})
        measures = pd.DataFrame(measures,
                                index=['N', 'mode', 'median', 'mean',
                                       '25%', '75%', 'interquart. range',
                                       'interquart. range (norm.)',
                                       'min', 'max', 'range', 'std', 'var',
                                       'entropy coef.', 'quality var.'])
        if len(self._mappers) > 0:
            for var in self._mappers.keys():
                measures.loc[['mode', 'median', '25%', '75%', 'min', 'max'], var] = \
                measures.loc[['mode', 'median', '25%', '75%', 'min', 'max'], var].map(self._mappers[var][1])
        return measures.T
    
    @staticmethod
    def _entropy_coef(series):
        p = series.value_counts(normalize=True) 
        entr_obs = (p * p.apply(np.log)).sum()
        n = len(p)
        p_exp = np.array([1/n] * n)
        entr_exp = (p_exp * np.log(p_exp)).sum()
        coef = entr_obs / entr_exp
        return coef
    
    @staticmethod
    def _cqv_coef(series):
        n = len(series)
        p = series.value_counts()
        k = len(p)
        p_sq_sum = (p ** 2).sum()
        coef = (k * (n**2 - p_sq_sum)) / ((n**2) * (k - 1))
        return coef
    
    def summary(self):
        """
        Return aggregated results of the analysis.
        """
        return self._get_statistics()
    
    def __repr__(self):
        n_vars_ = len(self._variables)
        return f'<ScaleStatistics Object for {n_vars_} variables>'
    
    @staticmethod
    def _get_frequencies(series):
        raw = series.value_counts()
        raw.name = 'N'
        normalized = series.value_counts(normalize=True) * 100
        normalized.name = '%'
        return pd.concat([raw, normalized], axis=1)
    
    def frequencies(self):
        """
        Return a dictionary of all frequency tables. 
        To get a particular frequency table, use a variable's name as a key of the dictionary. 
        """
        freq = {}
        for var in self._variables:
            ser = self._data[var]
            freqs = OrdinalStatistics._get_frequencies(ser)
            if var in self._mappers:
                freqs.index = freqs.index.map(self._mappers[var][1])
            freq.update({var: freqs})
        return freq
